##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
__version__ = "1.1.1"

import importlib.util
import os
import platform
import sys

_loaded_cpp_module = None


def load_cpp_module():
    global _loaded_cpp_module
    module_name = "gbrl_cpp"
    python_version = (f"cpython-{sys.version_info.major}"
                      f"{sys.version_info.minor}")
    python_version_short = (f"cp{sys.version_info.major}"
                            f"{sys.version_info.minor}")

    if platform.system() == "Windows":
        ext = ".pyd"
    elif platform.system() == "Darwin":  # macOS
        ext = ".dylib"
    else:  # Assume Linux/Unix
        ext = ".so"
    possible_paths = [
        os.path.join(os.path.dirname(__file__)),  # Current directory
        os.path.join(os.path.dirname(__file__), "Release"),  # Release folder
    ]

    for dir_path in possible_paths:
        if os.path.exists(dir_path):
            # Scan for files that match the module name and extension
            for file_name in os.listdir(dir_path):
                if file_name.startswith(module_name) and \
                    file_name.endswith(ext) and \
                    (python_version in file_name or
                        python_version_short in file_name):
                    # Dynamically load the matching shared library
                    file_path = os.path.join(dir_path, file_name)
                    spec = importlib.util.spec_from_file_location(module_name,
                                                                  file_path)
                    module = importlib.util.module_from_spec(spec)  # type: ignore
                    spec.loader.exec_module(module)  # type: ignore
                    _loaded_cpp_module = module = module
                    return module

    if platform.system() == "Darwin":  # check for .so on Darwin
        ext = ".so"
        for dir_path in possible_paths:
            if os.path.exists(dir_path):
                # Scan for files that match the module name and extension
                for file_name in os.listdir(dir_path):
                    if file_name.startswith(module_name) and \
                        file_name.endswith(ext) and (python_version in
                                                     file_name or
                                                     python_version_short
                                                     in file_name):
                        # Dynamically load the matching shared library
                        file_path = os.path.join(dir_path, file_name)
                        spec = importlib.util.spec_from_file_location(
                            module_name, file_path)
                        module = importlib.util.module_from_spec(spec)  # type: ignore
                        spec.loader.exec_module(module)  # type: ignore
                        _loaded_cpp_module = module = module
                        return module
    raise ImportError(f"Could not find {module_name}{ext} in any of the"
                      f"expected locations: {possible_paths}")


# Load the C++ module dynamically
_gbrl_cpp_module = load_cpp_module()

# Create a global alias for the GBRL class
GBRL_CPP = _gbrl_cpp_module.GBRL

cuda_available = GBRL_CPP.cuda_available



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
from abc import ABC, abstractmethod
from typing import Tuple, Optional, Union

import numpy as np
import torch as th

from gbrl.common.utils import NumericalData
from gbrl.learners.gbt_learner import GBTLearner


class BaseGBT(ABC):
    def __init__(self):
        """General class for gradient boosting trees
        """
        self.learner: Optional[GBTLearner] = None
        self.grad: Optional[NumericalData] = None
        self.params: Optional[NumericalData] = None
        self.input: Optional[NumericalData] = None

    def set_bias(self, *args, **kwargs) -> None:
        """Sets GBRL bias

        Args:
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)
        """
        raise NotImplementedError

    def set_feature_weights(self, feature_weights: NumericalData) -> None:
        """Sets GBRL feature weights

        Args:
            feature_weights (NumericalData): Feature weights to set
        """
        assert self.learner is not None, "Learner is not initialized"

        if isinstance(feature_weights, th.Tensor):
            feature_weights = feature_weights.clone().detach().cpu().numpy()
        # GBRL works with 2D numpy arrays.
        if len(feature_weights.shape) == 1:
            feature_weights = feature_weights[:, np.newaxis]
        self.learner.set_feature_weights(feature_weights.astype(np.single))

    def get_iteration(self) -> Union[int, Tuple[int, ...]]:
        """Gets the number of boosting iterations per learner

        Returns:
            Union[int, Tuple[int, ...]]: Number of boosting iterations per learner
        """
        assert self.learner is not None, "Learner is not initialized"
        return self.learner.get_iteration()

    def get_total_iterations(self) -> int:
        """Gets the total number of boosting iterations

        Returns:
            int: Total number of boosting iterations
                (sum of actor and critic if they are not shared, otherwise equals get_iteration())
        """
        assert self.learner is not None, "Learner is not initialized"
        return self.learner.get_total_iterations()

    def get_schedule_learning_rates(self) -> Union[float, Tuple[float, ...]]:
        """Gets learning rate values for optimizers according to schedule of ensemble

        Constant schedule - no change in values.
        Linear schedule - learning rate value according to number of trees in the ensemble.

        Returns:
            Union[float, Tuple[float, ...]]: Learning rate schedule per optimizer
        """
        assert self.learner is not None, "Learner is not initialized"
        return self.learner.get_schedule_learning_rates()

    @abstractmethod
    def step(self, *args, **kwargs) -> None:
        """Performs a boosting step (fits a single tree on the gradients)

        Args:
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)
        """
        pass

    def fit(self, *args, **kwargs) -> Union[float, Tuple[float, ...]]:
        """Fits multiple iterations (as in supervised learning)

        Args:
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)

        Returns:
            Union[float, Tuple[float, ...]]: Final loss per learner over all examples
        """
        raise NotImplementedError

    def get_num_trees(self, *args, **kwargs) -> Union[int, Tuple[int, ...]]:
        """Gets the number of trees in the ensemble per learner

        Args:
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)

        Returns:
            Union[int, Tuple[int, ...]]: Number of trees in the ensemble per learner
        """
        assert self.learner is not None, "Learner is not initialized"
        return self.learner.get_num_trees(*args, **kwargs)

    def tree_shap(self, tree_idx: int, features: NumericalData, *args, **kwargs) -> \
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Calculates SHAP values for a single tree

        Implementation based on - https://github.com/yupbank/linear_tree_shap.
        See Linear TreeShap, Yu et al, 2023, https://arxiv.org/pdf/2209.08192.

        Args:
            tree_idx (int): Tree index
            features (NumericalData): Input features
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]: SHAP values of shape
                [n_samples, number of input features, number of outputs]. The output
                is a tuple of SHAP values per model only in the case of a separate
                actor-critic model.
        """
        assert self.learner is not None, "Learner is not initialized"
        return self.learner.tree_shap(tree_idx, features, *args, **kwargs)

    def shap(self, features: NumericalData, *args, **kwargs) -> \
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Calculates SHAP values for the entire ensemble

        Implementation based on - https://github.com/yupbank/linear_tree_shap.
        See Linear TreeShap, Yu et al, 2023, https://arxiv.org/pdf/2209.08192.

        Args:
            features (NumericalData): Input features
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]: SHAP values of shape
                [n_samples, number of input features, number of outputs]. The output
                is a tuple of SHAP values per model only in the case of a separate
                actor-critic model.
        """
        assert self.learner is not None, "Learner is not initialized"
        return self.learner.shap(features, *args, **kwargs)

    def save_learner(self, save_path: str) -> None:
        """Saves model to file

        Args:
            save_path (str): Absolute path and name of save filename
        """
        assert self.learner is not None, "Learner is not initialized"
        self.learner.save(save_path)

    def export_learner(self, filename: str, modelname: Optional[str] = None) -> None:
        """Exports learner model as a C-header file

        Args:
            filename (str): Absolute path and name of exported filename
            modelname (Optional[str], optional): Model name for export. Defaults to None.
        """
        assert self.learner is not None, "Learner is not initialized"
        self.learner.export(filename, modelname)

    @classmethod
    def load_learner(cls, load_name: str, device: str) -> "BaseGBT":
        """Loads a BaseGBT model from a file

        Args:
            load_name (str): Full path to the saved model file
            device (str): Device to load the model onto ('cpu' or 'cuda')

        Returns:
            BaseGBT: Loaded BaseGBT model instance
        """
        instance = cls.__new__(cls)
        instance.learner = GBTLearner.load(load_name, device)
        instance.grad = None
        instance.params = None
        instance.input = None
        return instance

    def get_params(self) -> Tuple[Union[NumericalData, Tuple[NumericalData, ...]],
                                  Optional[Union[NumericalData, Tuple[NumericalData, ...]]]]:
        """Gets predicted model parameters and their respective gradients

        Returns:
            Tuple[Union[NumericalData, Tuple[NumericalData, ...]],
                  Optional[Union[NumericalData, Tuple[NumericalData, ...]]]]:
                Model parameters and their gradients
        """
        assert self.params is not None, "must run a forward pass first"
        params = self.params
        if isinstance(self.params, tuple):
            params = (params[0].detach().cpu().numpy(), params[1].detach().cpu().numpy())
        return params, self.grad

    def set_device(self, device: str):
        """Sets GBRL device (either cpu or cuda)

        Args:
            device (str): Device choice, must be in ['cpu', 'cuda']
        """
        assert device in ['cpu', 'cuda'], "device must be in ['cpu', 'cuda']"
        assert self.learner is not None, "Learner is not initialized"
        self.learner.set_device(device)

    def get_device(self) -> Union[str, Tuple[str, str]]:
        """Gets GBRL device/devices per learner

        Returns:
            Union[str, Tuple[str, str]]: GBRL device per model
        """
        assert self.learner is not None, "Learner is not initialized"
        return self.learner.get_device()

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Union[NumericalData, Tuple[NumericalData, ...]]:
        """Returns GBRL's output as either a Tensor or a numpy array per learner

        Args:
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)

        Returns:
            Union[NumericalData, Tuple[NumericalData, ...]]: Model output per learner
        """
        pass

    def print_tree(self, tree_idx: int, *args, **kwargs) -> None:
        """Prints tree information

        Args:
            tree_idx (int): Tree index to print
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)
        """
        assert self.learner is not None, "Learner is not initialized"
        self.learner.print_tree(tree_idx, *args, **kwargs)

    def plot_tree(self, tree_idx: int, filename: str, *args, **kwargs) -> None:
        """Plots tree using graphviz (only works if GBRL was compiled with graphviz)

        Args:
            tree_idx (int): Tree index to plot
            filename (str): .png filename to save
            *args: Variable length argument list (implementation dependent)
            **kwargs: Arbitrary keyword arguments (implementation dependent)
        """
        assert self.learner is not None, "Learner is not initialized"
        self.learner.plot_tree(tree_idx, filename, *args, **kwargs)

    def copy(self) -> "BaseGBT":
        """Creates a copy of the class instance

        Returns:
            BaseGBT: Copy of the current instance
        """
        return self.__copy__()

    @abstractmethod
    def __copy__(self) -> "BaseGBT":
        """Copy constructor

        Returns:
            BaseGBT: Copy of the current instance
        """
        pass



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
import os
from typing import Dict, Optional, Tuple, Union

import numpy as np
from gbrl.common.utils import NumericalData

from gbrl.learners.actor_critic_learner import (SeparateActorCriticLearner,
                                                SharedActorCriticLearner)
from gbrl.models.base import BaseGBT
from gbrl.common.utils import (clip_grad_norm, numerical_dtype, setup_optimizer,
                               validate_array)


class ActorCritic(BaseGBT):
    """
    GBRL model for a shared Actor and Critic ensemble.

    Supports both shared and separate actor-critic tree structures.
    """
    def __init__(self,
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int,
                 policy_optimizer: Dict,
                 value_optimizer: Dict,
                 shared_tree_struct: bool = True,
                 params: Dict = dict(),
                 bias: Optional[np.ndarray] = None,
                 verbose: int = 0,
                 device: str = 'cpu'):
        """
        GBRL model for a shared Actor and Critic ensemble.

        Args:
         tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
        output_dim (int): output dimension.
        policy_optimizer Dict: dictionary containing policy optimizer
        parameters (see GBRL class for optimizer details).
        value_optimizer Dict: dictionary containing value optimizer parameters
        (see GBRL class for optimizer details).
        shared_tree_struct (bool, optional): sharing actor and critic.
        Defaults to True.
        params (Dict, optional): GBRL parameters such as:
            control_variates (bool): use control variates (variance reduction technique CPU only).
            split_score_func (str): "cosine" or "l2"
            generator_type- (str): candidate generation method "Quantile" or "Uniform".
            feature_weights - (list[float]): Per-feature multiplication
            weights used when choosing the best split. Weights should be >= 0
        bias (np.ndarray, optional): manually set a bias. Defaults to None = np.zeros.
        verbose (int, optional): verbosity level. Defaults to 0.
        device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """
        assert output_dim > 1, \
            "ActorCritic module requires at least 1 output dimension each for the actor and the critic."
        super().__init__()

        policy_optimizer = setup_optimizer(policy_optimizer, prefix='policy_')
        value_optimizer = setup_optimizer(value_optimizer, prefix='value_')

        self.shared_tree_struct = True if value_optimizer is None else \
            shared_tree_struct
        bias = bias if bias is not None else np.zeros(output_dim if
                                                      shared_tree_struct
                                                      else output_dim - 1,
                                                      dtype=numerical_dtype)
        # init model
        if self.shared_tree_struct:
            self.learner = SharedActorCriticLearner(input_dim, output_dim,
                                                    tree_struct,
                                                    policy_optimizer,
                                                    value_optimizer,
                                                    params, verbose, device)
            self.learner.reset()
            self.learner.set_bias(bias)
        else:
            self.learner = SeparateActorCriticLearner(input_dim, output_dim,  # type: ignore
                                                      tree_struct,
                                                      policy_optimizer,
                                                      value_optimizer,
                                                      params, verbose, device)
            self.learner.reset()  # type: ignore
            self.learner.set_bias(bias, model_idx=0)  # type: ignore
        self.policy_grad = None
        self.value_grad = None

    def set_bias(self, bias: Union[np.ndarray, float], model_idx: Optional[int] = None) -> None:
        """Sets the bias for the actor or critic model.

        Args:
            bias (Union[np.ndarray, float]): The bias value(s) to set.
            model_idx (int): The index of the model (0 for actor, 1 for critic).
        """
        assert self.learner is not None, "Learner is not initialized"
        assert model_idx is None or isinstance(self.learner, SeparateActorCriticLearner), \
            "Cannot specify model index for a SharedActorCriticLearner."

        if model_idx is None:
            assert isinstance(bias, np.ndarray), \
                "Model idx not specified, bias is expected to be a numpy array with same dimension as the output dim."
            if isinstance(self.learner, SharedActorCriticLearner):
                self.learner.set_bias(bias)
            else:
                self.learner.set_bias(bias[:-1], model_idx=0)  # type: ignore
                self.learner.set_bias(bias[-1], model_idx=1)  # type: ignore

        self.learner.set_bias(bias, model_idx=model_idx)  # type: ignore

    @classmethod
    def load_learner(cls, load_name: str, device: str) -> "ActorCritic":
        """Loads GBRL model from a file

        Args:
            load_name (str): full path to file name

        Returns:
            ActorCritic: loaded ActorCriticModel
        """
        policy_file = load_name + '_policy.gbrl_model'
        value_file = load_name + '_value.gbrl_model'

        instance = cls.__new__(cls)
        if os.path.isfile(policy_file) and os.path.isfile(value_file):
            instance.learner = SeparateActorCriticLearner.load(load_name, device)  # type: ignore
            instance.shared_tree_struct = False
        else:
            instance.learner = SharedActorCriticLearner.load(load_name, device)  # type: ignore
            instance.shared_tree_struct = True

        instance.policy_grad = None
        instance.value_grad = None
        instance.params = None
        instance.input = None
        return instance

    def predict_values(self, observations: NumericalData,
                       requires_grad: bool = True, start_idx: int = 0,
                       stop_idx: Optional[int] = None, tensor: bool = True) -> \
            NumericalData:
        """
        Predict only values. If `requires_grad=True` then stores differentiable parameters in self.params
           Return type/device is identical to the input type/device.

        Args:
            observations (NumericalData)
            requires_grad (bool, optional). Defaults to True. Ignored if input is a numpy array.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses
            all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            NumericalData: values
        """
        assert self.learner is not None, "Learner is not initialized"

        values = self.learner.predict_critic(observations, requires_grad,  # type: ignore
                                             start_idx, stop_idx, tensor)
        if requires_grad:
            self.value_grad = None
            self.params = values
        return values

    def __call__(self, observations: NumericalData,
                 requires_grad: bool = True, start_idx: int = 0,
                 stop_idx: Optional[int] = None, tensor: bool = True) -> \
            Tuple[NumericalData, NumericalData]:
        """
        Predicts  and returns actor and critic outputs as tensors.
        If `requires_grad=True` then stores differentiable parameters in self.params
           Return type/device is identical to the input type/device.
        Args:
            observations (NumericalData)
            requires_grad (bool, optional). Defaults to True. Ignored if input is a numpy array.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses
            all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Tuple[NumericalData, NumericalData]: actor and critic output
        """
        assert self.learner is not None, "Learner is not initialized"

        params = self.learner.predict(observations, requires_grad, start_idx,
                                      stop_idx, tensor)
        if requires_grad:
            self.policy_grad = None
            self.value_grad = None
            self.params = params
            self.input = observations
        return params  # type: ignore

    def step(self, observations: Optional[NumericalData] = None,
             policy_grad: Optional[NumericalData] = None,
             value_grad: Optional[NumericalData] = None,
             policy_grad_clip: Optional[float] = None,
             value_grad_clip: Optional[float] = None) -> None:
        """
        Performs a boosting step for both the actor and critic.

        If `observations` is not provided, it uses the stored input from thelast forward pass.

        Args:
            observations (Optional[NumericalData], optional):Input observations.
            policy_grad (Optional[NumericalData], optional):Manually computed gradients for the policy.
            value_grad (Optional[NumericalData], optional): Manually computed gradients for the value function.
            policy_grad_clip (Optional[float], optional):Gradient clipping value for policy updates.
            value_grad_clip (Optional[float], optional): Gradient clipping value for value updates.
        """
        assert self.learner is not None, "Learner is not initialized"
        if observations is None:
            assert self.input is not None, ("Cannot update trees without input."
                                            "Make sure model is called with requires_grad=True")
            observations = self.input
        n_samples = len(observations)

        policy_grad = policy_grad if policy_grad is not None else self.params[0].grad.detach() * n_samples  # type: ignore
        value_grad = value_grad if value_grad is not None else self.params[1].grad.detach() * n_samples  # type: ignore

        policy_grad = clip_grad_norm(policy_grad, policy_grad_clip)  # type: ignore
        value_grad = clip_grad_norm(value_grad, value_grad_clip)  # type: ignore

        validate_array(policy_grad)
        validate_array(value_grad)

        self.learner.step(observations, policy_grad, value_grad)  # type: ignore
        self.policy_grad = policy_grad
        self.value_grad = value_grad
        self.input = None

    def actor_step(self, observations: Optional[NumericalData]
                   = None, policy_grad: Optional[NumericalData]
                   = None, policy_grad_clip: Optional[float] = None) -> None:
        """
        Performs a single boosting step for the actor (should only be used
        if actor and critic use separate models)

        Args:
            observations (NumericalData):
            policy_grad_clip (float, optional): Defaults to None.
            policy_grad (Optional[NumericalData], optional): manually calculated gradients. Defaults to None.

        Returns:
            np.ndarray: policy gradient
        """
        assert self.learner is not None, "Learner is not initialized"
        assert not self.shared_tree_struct, "Cannot separate boosting steps"
        "for actor and critic when using separate tree architectures!"
        if observations is None:
            assert self.input is not None, ("Cannot update trees without input."
                                            "Make sure model is called with requires_grad=True")
            observations = self.input
        n_samples = len(observations)
        policy_grad = policy_grad if policy_grad is not None else self.params[0].grad.detach() * n_samples  # type: ignore
        policy_grad = clip_grad_norm(policy_grad, policy_grad_clip)  # type: ignore
        validate_array(policy_grad)

        self.learner.step_actor(observations, policy_grad)  # type: ignore
        self.policy_grad = policy_grad

    def critic_step(self, observations: Optional[NumericalData] = None,
                    value_grad: Optional[NumericalData] = None,
                    value_grad_clip: Optional[float] = None) -> None:
        """
        Performs a single boosting step for the critic (should only be used
        if actor and critic use separate models)

        Args:
            observations (NumericalData):
            value_grad_clip (float, optional): Defaults to None.
            value_grad (Optional[NumericalData], optional): manually calculated gradients. Defaults to None.

        Returns:
            np.ndarray: value gradient
        """
        assert self.learner is not None, "Learner is not initialized"
        assert not self.shared_tree_struct, ("Cannot separate boosting steps"
                                             "for actor and critic when using separate tree architectures!")
        if observations is None:
            assert self.input is not None, ("Cannot update trees without input."
                                            "Make sure model is called with requires_grad=True")
            observations = self.input
        n_samples = len(observations)

        value_grad = value_grad if value_grad is not None else self.params[1].grad.detach() * n_samples  # type: ignore
        value_grad = clip_grad_norm(value_grad, value_grad_clip)  # type: ignore

        validate_array(value_grad)
        self.learner.step_critic(observations, value_grad)  # type: ignore
        self.value_grad = value_grad

    def get_params(self) -> Tuple[NumericalData, NumericalData]:
        """
        Returns the predicted actor and critic parameters along with their gradients.

        Returns:
            Tuple[NumericalData, NumericalData]:
                - Predicted actor and critic outputs.
                - Corresponding policy and value gradients.
        """
        assert self.params is not None, "must run a forward pass first"
        if isinstance(self.params, tuple):
            params = (self.params[0].detach().cpu().numpy(), self.params[1].detach().cpu().numpy())
        else:
            params = self.params
        return params, (self.policy_grad, self.value_grad)  # type: ignore

    def save_learner(self, save_path: str) -> None:
        """
        Saves model to file

        Args:
            filename (str): Absolute path and name of save filename.
        """
        assert self.learner is not None, "Learner is not initialized"

        if self.shared_tree_struct:
            self.learner.save(save_path)
        else:
            self.learner.save(save_path, custom_names=['policy', 'value'])  # type: ignore

    def copy(self) -> "ActorCritic":
        """
        Copy class instance

        Returns:
            ActorCritic: copy of current instance
        """
        return self.__copy__()

    def __copy__(self) -> "ActorCritic":
        assert self.learner is not None, "Learner is not initialized"

        learner = self.learner.copy()
        copy_ = ActorCritic(learner.tree_struct, learner.input_dim,
                            learner.output_dim, learner.policy_optimizer,  # type: ignore
                            learner.value_optimizer, self.shared_tree_struct,   # type: ignore
                            learner.params, learner.get_bias(),  # type: ignore
                            learner.verbose, learner.device)
        copy_.learner = learner  # type: ignore
        return copy_



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch as th

from gbrl.common.utils import NumericalData, to_numpy


class BaseLearner(ABC):
    """
    Abstract base class for gradient boosting tree learners.

    This class defines the fundamental interface for gradient boosting
    learners and serves as a wrapper for the C++ backend.

    Attributes:
        input_dim (int): The number of input features.
        output_dim (int): The number of output dimensions.
        tree_struct (Dict): Dictionary containing tree structure parameters.
        params (Union[Dict, List[Dict]]): Dictionary containing model parameters.
        verbose (int): Verbosity level (0 = silent, 1 = debug).
        device (str): The device the model runs on (e.g., 'cpu' or 'cuda').
        iteration (int): The current training iteration.
        total_iterations (int): Total number of training iterations.
        feature_weights (np.ndarray): Feature importance weights.
    """
    def __init__(self, input_dim: int, output_dim: int, tree_struct: Dict,
                 params: Dict, verbose: int = 0, device: str = 'cpu'):
        """
        Initializes the BaseLearner.

        Args:
            input_dim (int): Number of input features.
            output_dim (int): Number of output dimensions.
            tree_struct (Dict): Dictionary containing tree
            structure parameters.
            params (Dict): Dictionary containing additional model parameters.
            verbose (int, optional): Verbosity level (0 = silent, 1 = debug).
            Defaults to 0.
            device (str, optional): Device to run the model on
            ('cpu' or 'cuda'). Defaults to 'cpu'.
        """
        self.tree_struct = tree_struct
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.device = device
        self.params = {'input_dim': input_dim,
                       'output_dim': output_dim,
                       'split_score_func': params.get('split_score_func',
                                                      'Cosine'),
                       'generator_type': params.get('generator_type',
                                                    'Quantile'),
                       'use_control_variates': params.get('control_variates',
                                                          False),
                       'verbose': verbose, 'device': device, **tree_struct}

        self.iteration = 0
        self.total_iterations = 0
        self.verbose = verbose
        feature_weights = params.get('feature_weights', None)
        if feature_weights is not None:
            feature_weights = to_numpy(feature_weights)
            feature_weights = feature_weights.flatten()
            message = "feature weights contains non-positive values"
            assert np.all(feature_weights >= 0), message
        else:
            weights = np.ones(input_dim, dtype=np.single)
            feature_weights = np.ascontiguousarray(weights)
        self.feature_weights = feature_weights
        self.cpp_model = None  # Placeholder for the C++ model instance

    @abstractmethod
    def reset(self) -> None:
        """Resets the model, reinitializing internal states and parameters."""
        pass

    @abstractmethod
    def step(self, *args, **kwargs) -> None:
        """Performs a single update step using provided gradients."""
        pass

    @abstractmethod
    def fit(self, *args, **kwargs) -> Union[float, List[float]]:
        """
        Trains the model on provided data.

        Returns:
            Union[float, List[float]]: The final loss after training per model.
        """
        pass

    @abstractmethod
    def save(self, filename: str, *args, **kwargs) -> None:
        """
        Saves the model to a file.

        Args:
            filename (str): The filename to save the model to.
        """
        pass

    def export(self, filename: str, modelname: Optional[str] = None) -> None:
        # exports model to C
        filename = filename.rstrip('.')
        filename += '.h'
        assert self.cpp_model is not None, "Can't export non-existent model!"
        if modelname is None:
            modelname = ""
        try:
            status = self.cpp_model.export(filename, modelname)
            assert status == 0, "Failed to export model"
        except RuntimeError as e:
            print(f"Caught an exception in GBRL: {e}")

    @classmethod
    @abstractmethod
    def load(cls, filename: str, device: str, *args, **kwargs) -> "BaseLearner":
        """
        Loads a model from a file.

        Args:
            filename (str): Path to the model file.
            device (str): Device to load the model onto.

        Returns:
            BaseLearner: Loaded model instance.
        """
        pass

    @abstractmethod
    def get_schedule_learning_rates(self, *args, **kwargs) -> Union[int, Tuple[int, ...]]:
        """
        Retrieves the scheduled learning rates.

        Returns:
            Union[int, Tuple[int, ...]]: Learning rate(s).
        """
        pass

    def get_total_iterations(self) -> int:
        """
        Returns the total number of iterations performed.

        Returns:
            int: The total number of iterations.
        """
        return self.total_iterations

    @abstractmethod
    def get_iteration(self, *args, **kwargs) -> Union[int, Tuple[int, ...]]:
        """
        Retrieves the current iteration count.

        Returns:
            Union[int, Tuple[int, ...]]: The current iteration number.
        """
        pass

    @abstractmethod
    def get_num_trees(self, *args, **kwargs) -> Union[int, Tuple[int, ...]]:
        """
        Retrieves the number of trees in the model.

        Returns:
            Union[int, Tuple[int, ...]]: Number of trees.
        """
        pass

    @abstractmethod
    def set_bias(self, bias: Union[np.ndarray, float], *args, **kwargs) -> None:
        """
        Sets the bias term for the model.

        Args:
            bias (Union[np.ndarray, float]): Bias value(s).
        """
        pass

    @abstractmethod
    def set_feature_weights(self, feature_weights: Union[np.ndarray, float], *args, **kwargs) -> None:
        """
        Sets the feature importance weights.

        Args:
            feature_weights (Union[np.ndarray, float]): Feature weights.
        """
        pass

    @abstractmethod
    def get_bias(self, *args, **kwargs) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
        """
        Retrieves the model bias.

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, ...]]: Bias values.
        """
        pass

    @abstractmethod
    def get_feature_weights(self, *args, **kwargs) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
        """
        Retrieves the feature importance weights.

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, ...]]: Feature weights.
        """
        pass

    @abstractmethod
    def get_device(self, *args, **kwargs) -> Union[str, Tuple[str, ...]]:
        """
        Retrieves the current device the model is running on.

        Returns:
            Union[str, Tuple[str, ...]]: Device (e.g., 'cpu', 'cuda').
        """
        pass

    @abstractmethod
    def print_tree(self, tree_idx: int, *args, **kwargs) -> None:
        """
        Prints the structure of a specific decision tree.

        Args:
            tree_idx (int): Index of the tree.
        """
        pass

    @abstractmethod
    def plot_tree(self, tree_idx: int, filename: str, *args, **kwargs) -> None:
        """
        Plots a decision tree and saves it to a file.

        Args:
            tree_idx (int): Index of the tree.
            filename (str): Path to save the tree visualization.
        """
        pass

    @abstractmethod
    def tree_shap(self, tree_idx: int, features: NumericalData, *args, **kwargs) -> \
            Union[np.ndarray, Tuple[np.ndarray, ...]]:
        """
        Computes SHAP values for a specific tree.

        Implementation based on - https://github.com/yupbank/linear_tree_shap
        See Linear TreeShap, Yu et al, 2023, https://arxiv.org/pdf/2209.08192
        Args:
            tree_idx (int): tree index
            features (NumericalData):

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, ...]]: shap values
        """
        pass

    @abstractmethod
    def shap(self, features: NumericalData, *args, **kwargs) -> \
            Union[np.ndarray, Tuple[np.ndarray, ...]]:
        """
        Computes SHAP values for the entire model.

        Uses Linear tree shap for each tree in the ensemble (sequentially)
        Implementation based on - https://github.com/yupbank/linear_tree_shap
        See Linear TreeShap, Yu et al, 2023, https://arxiv.org/pdf/2209.08192
        Args:
            features (NumericalData):

        Returns:
            Union[np.ndarray, Tuple[np.ndarray, ...]]: shap values
        """
        pass

    @abstractmethod
    def set_device(self, device: Union[str, th.device], *args, **kwargs) -> None:
        """
        Sets the device the model should run on.

        Args:
            device (Union[str, th.device]): Target device.
        """
        pass

    @abstractmethod
    def predict(self, *args, **kwargs) -> Union[np.ndarray, Tuple[np.ndarray, ...]]:
        """
        Generates predictions using the trained model.

        Returns:
            np.ndarray: Model predictions.
        """
        pass

    @abstractmethod
    def distil(self, *args, **kwargs) -> Tuple[int, Dict]:
        """
        Distills the model into a smaller, simplified version.

        Returns:
            Tuple[int, Dict]: Final loss and updated parameters.
        """
        pass

    def copy(self):
        """
        Creates a copy of the model instance.

        Returns:
            BaseLearner: A copy of the model.
        """
        return self.__copy__()

    @abstractmethod
    def print_ensemble_metadata(self) -> None:
        """Prints metadata information about the entire ensemble."""
        pass

    @abstractmethod
    def __copy__(self) -> "BaseLearner":
        """Creates and returns a copy of the learner instance."""
        pass



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
import os
import shutil
import sys
import tempfile
import unittest
from pathlib import Path

import numpy as np
import shap
import torch as th
from sklearn import datasets
from sklearn.tree import DecisionTreeRegressor
from torch.nn.functional import mse_loss

ROOT_PATH = Path(__file__).parent.parent
sys.path.insert(0, str(ROOT_PATH))

from gbrl import cuda_available
from gbrl.models.gbt import GBTModel
from tests import CATEGORICAL_INPUTS, CATEGORICAL_OUTPUTS

N_EPOCHS = 100


def rmse_model(model, X, y, n_epochs, device='cpu'):
    y_ = th.tensor(y, dtype=th.float32, device=device).squeeze()
    X_ = X.copy()
    epoch = 0
    while epoch < n_epochs:
        y_pred = model(X_, requires_grad=True)
        loss = 0.5*mse_loss(y_pred, y_)
        loss.backward()
        model.step()
        print(f"epoch: {epoch} loss: {loss.sqrt()}")
        epoch += 1
    y_pred = model(X_)
    loss = (0.5*mse_loss(y_pred, y_)).sqrt().item()
    return loss


def to_utf8(s):
    return s.encode('utf-8').decode('utf-8') if isinstance(s, str) else s


class TestGBTSingle(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print('Loading data...')
        # Imagine this loads your actual data
        try:
            X, y = datasets.load_diabetes(return_X_y=True,
                                          as_frame=False, scaled=False)
        except TypeError:  # python3.7 uses older version of scikit-learn
            X, y = datasets.load_diabetes(return_X_y=True, as_frame=False)
        out_dim = 1 if len(y.shape) == 1 else y.shape[1]
        if out_dim == 1:
            y = y[:, np.newaxis]
        input_dim = X.shape[1]
        cls.single_data = (X, y)
        cls.out_dim = out_dim
        cls.input_dim = input_dim
        cls.n_epochs = 100
        cls.test_dir = tempfile.mkdtemp()
        # Setup categorical data
        X_categorical = np.array(CATEGORICAL_INPUTS, dtype=str)
        X_categorical = np.char.encode(X_categorical, encoding='utf-8',
                                       errors=None)
        X_categorical = np.char.decode(X_categorical, encoding='utf-8',
                                       errors=None)
        y_categorical = np.array(CATEGORICAL_OUTPUTS,
                                 dtype=np.single)[:, np.newaxis]
        cls.cat_data = (X_categorical, y_categorical)
        cls.tree_struct = {'max_depth': 4,
                           'n_bins': 256, 'min_data_in_leaf': 0,
                           'par_th': 2,
                           'grow_policy': 'greedy'}
        cls.sgd_optimizer = {'algo': 'SGD',
                             'lr': 1.0,
                             'start_idx': 0,
                             'stop_idx': out_dim
                             }

    @classmethod
    def tearDownClass(cls):
        # Remove the directory after the test
        shutil.rmtree(cls.test_dir)

    def test_cosine_cpu(self):
        print("Running test_cosine_cpu")
        X, y = self.single_data
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine",
                       "generator_type": "Quantile"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        value = 5
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir, 'test_cosine_cpu'))

        model.learner.reset()
        model.set_bias_from_targets(y)
        train_loss = model.fit(X, y, self.n_epochs)
        self.assertTrue(train_loss < value, f'Expected loss = {train_loss} '
                        f'< {value}')

        X_categorical, y_categorical = self.cat_data
        model = GBTModel(input_dim=X_categorical.shape[1],
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X_categorical, y_categorical, self.n_epochs)
        value = 5000
        self.assertTrue(loss < value, f'Expected Categorical loss = '
                        f'{loss} < {value}')

    def test_copy_cpu(self):
        print("Running test_copy_cpu")
        X, y = self.single_data
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine",
                       "generator_type": "Quantile"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        _ = rmse_model(model, X, y, self.n_epochs)
        copy_model = model.copy()
        y_pred = model(X, requires_grad=False, tensor=False)
        y_copy_pred = copy_model(X, requires_grad=False, tensor=False)
        assert np.allclose(y_pred, y_copy_pred), "Expected copied GBRL model "
        "to be equal to original"

    def test_continuation_cpu(self):
        print("Running test_continuation_cpu")
        X, y = self.single_data
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine",
                       "generator_type": "Quantile"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs // 2)
        model.save_learner(os.path.join(self.test_dir,
                                        'test_continuation_cpu'))
        new_model = GBTModel.load_learner(os.path.join(self.test_dir,
                                                       'test_continuation_cpu'),
                                          device='cpu')
        loss = rmse_model(new_model, X, y, self.n_epochs // 2)
        value = 5
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')

    @unittest.skipIf(not cuda_available(), "cuda not available skipping over "
                     "gpu tests")
    def test_continuation_gpu(self):
        print("Running test_continuation_gpu")
        X, y = self.single_data
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine",
                       "generator_type": "Quantile"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs // 2)
        model.save_learner(os.path.join(self.test_dir,
                                        'test_continuation_gpu'))
        new_model = GBTModel.load_learner(os.path.join(self.test_dir,
                                                       'test_continuation_gpu'),
                                          device='cuda')
        loss = rmse_model(new_model, X, y, self.n_epochs // 2, device='cuda')
        value = 5
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')

    def test_shap_cpu(self):
        print("Running test_shap_cpu")
        X, y = self.single_data
        tree_struct = {'max_depth': 3,
                       'n_bins': 256, 'min_data_in_leaf': 1,
                       'par_th': 2,
                       'grow_policy': 'greedy'}
        params = dict({"control_variates": False,
                       "split_score_func": "L2",
                       "generator_type": "Uniform"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.learner.step(X, y)
        gbrl_shap = model.tree_shap(0, X[0, :]).flatten()
        clf = DecisionTreeRegressor(max_depth=3).fit(X, y)

        target_shap = shap.TreeExplainer(clf).shap_values(X[0])
        self.assertTrue(np.allclose(gbrl_shap, target_shap, rtol=1e-3),
                        'GBRL SHAP values are not close to target SHAP values')

    def test_cosine_adam_cpu(self):
        print("Running test_cosine_adam_cpu")
        X, y = self.single_data
        optimizer = {'algo': 'Adam',
                     'lr': 1.0,
                     'start_idx': 0,
                     'stop_idx': self.out_dim
                     }
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        value = 50
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir, 'test_cosine_adam_cpu'))

    @unittest.skipIf(not cuda_available(), "cuda not available skipping "
                     "over gpu tests")
    def test_cosine_gpu(self):
        print("Running test_cosine_gpu")
        X, y = self.single_data
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs, device='cuda')
        value = 2
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir, 'test_cosine_gpu'))

        model.learner.reset()
        model.set_bias_from_targets(y)
        train_loss = model.fit(X, y, self.n_epochs)
        value = 2
        self.assertTrue(train_loss < value, 'Expected loss = '
                        f'{train_loss} < {value}')

        X_categorical, y_categorical = self.cat_data
        model = GBTModel(input_dim=X_categorical.shape[1],
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y_categorical)
        loss = rmse_model(model, X_categorical, y_categorical,
                          self.n_epochs, device='cuda')
        value = 5000
        self.assertTrue(loss < value,
                        f'Expected Categorical loss = {loss} < {value}')

    def test_cosine_oblivious_cpu(self):
        print("Running test_cosine_oblivious_cpu")
        X, y = self.single_data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        value = 13
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir,
                                        'test_cosine_oblivious_gpu'))
        model.learner.reset()
        model.set_bias_from_targets(y)
        train_loss = model.fit(X, y, self.n_epochs)
        self.assertTrue(train_loss < value,
                        f'Expected loss = {train_loss} < {value}')
        X_categorical, y_categorical = self.cat_data
        model = GBTModel(input_dim=X_categorical.shape[1],
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y_categorical)
        loss = rmse_model(model, X_categorical, y_categorical, self.n_epochs)
        value = 5000
        self.assertTrue(loss < value,
                        f'Expected Categorical loss = {loss} < {value}')

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_cosine_oblivious_gpu(self):
        print("Running test_cosine_oblivious_gpu")
        X, y = self.single_data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs, device='cuda')
        value = 12
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir,
                                        'test_cosine_oblivious_gpu'))
        model.learner.reset()
        model.set_bias_from_targets(y)
        train_loss = model.fit(X, y, self.n_epochs)
        self.assertTrue(train_loss < value,
                        f'Expected loss = {train_loss} < {value}')

        X_categorical, y_categorical = self.cat_data
        model = GBTModel(input_dim=X_categorical.shape[1],
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y_categorical)
        loss = rmse_model(model, X_categorical, y_categorical, self.n_epochs,
                          device='cuda')
        value = 5000
        self.assertTrue(loss < value,
                        f'Expected Categorical loss = {loss} < {value}')

    def test_l2_cpu(self):
        print("Running test_l2_cpu")
        X, y = self.single_data
        params = dict({"control_variates": False, "split_score_func": "L2"})
        model = GBTModel(
                    input_dim=self.input_dim,
                    output_dim=self.out_dim,
                    tree_struct=self.tree_struct,
                    optimizers=self.sgd_optimizer,
                    params=params,
                    verbose=1,
                    device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        self.assertTrue(loss < 0.5, f'Expected loss = {loss} < 0.5')
        model.save_learner(os.path.join(self.test_dir, 'test_l2_cpu'))
        model.learner.reset()
        model.set_bias_from_targets(y)
        train_loss = model.fit(X, y, self.n_epochs)
        value = 6.0
        self.assertTrue(train_loss < value,
                        f'Expected loss = {train_loss} < {value}')
        X_categorical, y_categorical = self.cat_data
        model = GBTModel(input_dim=X_categorical.shape[1],
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y_categorical)
        loss = rmse_model(model, X_categorical, y_categorical, self.n_epochs)
        value = 5000
        self.assertTrue(loss < value,
                        f'Expected Categorical loss = {loss} < {value}')

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_l2_gpu(self):
        print("Running test_l2_gpu")
        X, y = self.single_data
        params = dict({"control_variates": False, "split_score_func": "L2"})
        model = GBTModel(
                    input_dim=self.input_dim,
                    output_dim=self.out_dim,
                    tree_struct=self.tree_struct,
                    optimizers=self.sgd_optimizer,
                    params=params,
                    verbose=0,
                    device='cuda')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs, device='cuda')

        self.assertTrue(loss < 0.5, f'Expected loss = {loss} < 0.5')
        model.save_learner(os.path.join(self.test_dir, 'test_l2_gpu'))
        model.learner.reset()
        model.set_bias_from_targets(y)
        train_loss = model.fit(X, y, self.n_epochs)
        value = 0.5
        self.assertTrue(train_loss < value,
                        f'Expected loss = {train_loss} < {value}')
        X_categorical, y_categorical = self.cat_data
        model = GBTModel(input_dim=X_categorical.shape[1],
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y_categorical)
        loss = rmse_model(model, X_categorical, y_categorical, self.n_epochs,
                          device='cuda')
        value = 5000
        self.assertTrue(loss < value,
                        f'Expected Categorical loss = {loss} < {value}')

    def test_l2_oblivious_cpu(self):
        print("Running test_l2_oblivious_cpu")
        X, y = self.single_data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False, "split_score_func": "L2"})
        model = GBTModel(
                    input_dim=self.input_dim,
                    output_dim=self.out_dim,
                    tree_struct=tree_struct,
                    optimizers=self.sgd_optimizer,
                    params=params,
                    verbose=0,
                    device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        self.assertTrue(loss < 10.0, f'Expected loss = {loss} < 10.0')
        model.save_learner(os.path.join(self.test_dir,
                                        'test_l2_oblivious_cpu'))

        model.learner.reset()
        model.set_bias_from_targets(y)
        train_loss = model.fit(X, y, self.n_epochs)
        value = 10.0
        self.assertTrue(train_loss < value,
                        f'Expected loss = {train_loss} < {value}')

        X_categorical, y_categorical = self.cat_data

        model = GBTModel(input_dim=X_categorical.shape[1],
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y_categorical)
        loss = rmse_model(model, X_categorical, y_categorical, self.n_epochs)
        value = 5000
        self.assertTrue(loss < value,
                        f'Expected Categorical loss = {loss} < {value}')

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_l2_oblivious_gpu(self):
        print("Running test_l2_oblivious_gpu")
        X, y = self.single_data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False, "split_score_func": "L2"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs, device='cuda')
        self.assertTrue(loss < 10.0, f'Expected loss = {loss} < 10.0')
        model.save_learner(os.path.join(self.test_dir,
                                        'test_l2_oblivious_gpu'))
        model.learner.reset()
        model.set_bias_from_targets(y)
        train_loss = model.fit(X, y, self.n_epochs)
        value = 10.0
        self.assertTrue(train_loss < value,
                        f'Expected loss = {train_loss} < {value}')
        X_categorical, y_categorical = self.cat_data
        model = GBTModel(input_dim=X_categorical.shape[1],
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y_categorical)
        loss = rmse_model(model, X_categorical, y_categorical,
                          self.n_epochs, device='cuda')
        value = 5000
        self.assertTrue(loss < value,
                        f'Expected Categorical loss = {loss} < {value}')

    def test_loading(self):
        X, y = self.single_data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False,
                       "split_score_func": "cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model = GBTModel.load_learner(os.path.join(self.test_dir,
                                                   'test_cosine_cpu'),
                                      device='cpu')
        y_pred = model(X, requires_grad=False, tensor=False)
        loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
        self.assertTrue(loss < 2.0, f'Expected loss = {loss} < 2.0')

        model = GBTModel.load_learner(os.path.join(self.test_dir, 'test_l2_cpu'),
                                      device='cpu')
        y_pred = model(X, requires_grad=False, tensor=False)
        loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
        self.assertTrue(loss < 0.5, f'Expected loss = {loss} < 0.5')
        if (cuda_available()):
            model = GBTModel.load_learner(os.path.join(self.test_dir,
                                                       'test_cosine_gpu'),
                                          device='cuda')
            y_pred = model(X, requires_grad=False, tensor=False)
            loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
            self.assertTrue(loss < 2.0, f'Expected loss = {loss} < 2.0')
            model = GBTModel.load_learner(
                os.path.join(self.test_dir,
                             'test_cosine_oblivious_gpu'), device='cuda')
            y_pred = model(X, requires_grad=False, tensor=False)
            loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
            self.assertTrue(loss < 12.0, f'Expected loss = {loss} < 12.0')
            model = GBTModel.load_learner(
                os.path.join(self.test_dir, 'test_l2_gpu'), device='cuda')
            y_pred = model(X, requires_grad=False, tensor=False)
            loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
            self.assertTrue(loss < 0.5, f'Expected loss = {loss} < 0.5')
            model = GBTModel.load_learner(os.path.join(self.test_dir,
                                                       'test_l2_oblivious_gpu'),
                                          device='cuda')
            y_pred = model(X, requires_grad=False, tensor=False)
            loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
            self.assertTrue(loss < 10.0, f'Expected loss = {loss} < 10.0')
        model = GBTModel.load_learner(os.path.join(self.test_dir,
                                                   'test_cosine_adam_cpu'),
                                      device='cpu')
        y_pred = model(X, requires_grad=False, tensor=False)
        loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
        value = 50.0
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')


if __name__ == '__main__':
    unittest.main()



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
from typing import Dict, List, Optional, Tuple

import numpy as np
import torch as th

from gbrl import GBRL_CPP
from gbrl.common.utils import (NumericalData, concatenate_arrays,
                               ensure_leaf_tensor_or_array, ensure_same_type,
                               get_tensor_info, numerical_dtype,
                               preprocess_features)
from gbrl.learners.gbt_learner import GBTLearner
from gbrl.learners.multi_gbt_learner import MultiGBTLearner


class SharedActorCriticLearner(GBTLearner):
    """
    SharedActorCriticLearner is a variant of GBTLearner where a single tree is
    used for both
    actor (policy) and critic (value) learning. It utilizes gradient boosting
    trees (GBTs)
    to estimate both policy and value function parameters efficiently.
    """
    def __init__(self, input_dim: int, output_dim: int, tree_struct: Dict,
                 policy_optimizer: Dict, value_optimizer: Dict,
                 params: Dict = dict(), verbose: int = 0, device: str = 'cpu'):
        """
        Initializes the SharedActorCriticLearner.

        Args:
            input_dim (int): Number of input features.
            output_dim (int): Number of output dimensions.
            tree_struct (Dict): Dictionary containing tree structure
            parameters.
            policy_optimizer (Dict): Dictionary with optimization parameters
            for the policy.
            value_optimizer (Dict): Dictionary with optimization parameters
            for the critic.
            params (Dict, optional): Additional model parameters. Defaults to
            an empty dictionary.
            verbose (int, optional): Verbosity level. Defaults to 0.
            device (str, optional): Device to run the model on. Defaults to
            'cpu'.
        """
        if verbose > 0:
            print('****************************************')
            print(f'Shared GBRL Tree with input dim: {input_dim},'
                  f'output dim: {output_dim}, tree_struct: {tree_struct}'
                  f'policy_optimizer: {policy_optimizer}'
                  f'value_optimizer: {value_optimizer}')
            print('****************************************')
        super().__init__(input_dim, output_dim, tree_struct,
                         [policy_optimizer, value_optimizer],
                         params, verbose, device)

    def step(self, obs: NumericalData,  # type: ignore
             theta_grad: np.ndarray, value_grad: np.ndarray) -> None:
        """
        Performs a gradient update step for both policy and value function.

        Args:
            obs (NumericalData): Input observations.
            theta_grad (np.ndarray): Gradient of the policy parameters.
            value_grad (np.ndarray): Gradient of the value function parameters.
        """
        assert self._cpp_model is not None, "Learner model not initialized"

        grads = concatenate_arrays(theta_grad, value_grad)
        obs, grads = ensure_same_type(obs, grads)
        if isinstance(obs, th.Tensor):
            obs = obs.float()
            grads = grads.float()  # type: ignore[attr-defined]
            # store data so that data isn't garbage collected
            # while GBRL uses it
            self._save_memory = (obs, grads)
            self._cpp_model.step(get_tensor_info(obs),
                                 None, get_tensor_info(grads))
            self._save_memory = None
        else:
            num_obs, cat_obs = preprocess_features(obs)
            grads = np.ascontiguousarray(grads).astype(numerical_dtype)
            input_dim = 0 if num_obs is None else num_obs.shape[1]
            input_dim += 0 if cat_obs is None else cat_obs.shape[1]
            self._cpp_model.step(num_obs, cat_obs, grads)

        self.iteration = self._cpp_model.get_iteration()  # type: ignore
        self.total_iterations += 1

    def distil(self, obs: NumericalData,  # type: ignore
               policy_targets: np.ndarray, value_targets: np.ndarray,
               params: Dict, verbose: int) -> Tuple[float, Dict]:
        """
        Distills the trained model into a student model.

        Args:
            obs (NumericalData): Input observations.
            policy_targets (np.ndarray): Target values for the policy (actor).
            value_targets (np.ndarray): Target values for the value function
            (critic).
            params (Dict): Distillation parameters.
            verbose (int): Verbosity level.

        Returns:
            Tuple[float, Dict]: The final loss value and updated parameters
            for distillation
        """
        targets = np.concatenate([policy_targets,
                                  value_targets[:, np.newaxis]], axis=1)
        return super().distil(obs, targets, params, verbose)

    def predict(self, obs: NumericalData,  # type: ignore
                requires_grad: bool = True, start_idx: int = 0,
                stop_idx: Optional[int] = None, tensor: bool = True) -> \
            Tuple[np.ndarray, np.ndarray]:
        """
        Predicts both policy and value function outputs.

        Args:
            obs (NumericalData): Input observations.
            requires_grad (bool, optional): Whether to compute gradients.
            Defaults to True.
            start_idx (int, optional): Start index for prediction. Defaults to
            0.
            stop_idx (int, optional): Stop index for prediction. Defaults to
            None.
            tensor (bool, optional): Whether to return a tensor. Defaults to
            True.

        Returns:
            Tuple[np.ndarray, np.ndarray]: Predicted policy and value outputs.
        """
        preds = super().predict(obs, requires_grad, start_idx, stop_idx,
                                tensor)
        pred_values = preds[:, -1]
        preds = preds[:, :-1]
        preds = ensure_leaf_tensor_or_array(preds, tensor, requires_grad, self.device)
        pred_values = ensure_leaf_tensor_or_array(pred_values, tensor, requires_grad, self.device)
        return preds, pred_values  # type: ignore

    def predict_policy(self, obs: NumericalData,
                       requires_grad: bool = True, start_idx: int = 0,
                       stop_idx: Optional[int] = None, tensor: bool = True):
        """
        Predicts the policy (actor) output for the given observations.

        Args:
            obs (NumericalData): Input observations.
            requires_grad (bool, optional): Whether to compute gradients.
            Defaults to True.
            start_idx (int, optional): Start index for prediction. Defaults to
            0.
            stop_idx (int, optional): Stop index for prediction. Defaults to
            None.
            tensor (bool, optional): Whether to return a tensor. Defaults to
            True.

        Returns:
            np.ndarray: Predicted policy outputs.
        """
        preds, _ = self.predict(obs, requires_grad, start_idx, stop_idx,
                                tensor)
        return preds

    def predict_critic(self, obs: NumericalData,
                       requires_grad: bool = True, start_idx: int = 0,
                       stop_idx: Optional[int] = None, tensor: bool = True):
        """
        Predicts the value function (critic) output for the given observations.

        Args:
            obs (NumericalData): Input observations.
            requires_grad (bool, optional): Whether to compute gradients. Defaults to True.
            start_idx (int, optional): Start index for prediction. Defaults to 0.
            stop_idx (int, optional): Stop index for prediction. Defaults to None.
            tensor (bool, optional): Whether to return a tensor. Defaults to True.

        Returns:
            np.ndarray: Predicted value function outputs.
        """
        _, pred_values = self.predict(obs, requires_grad, start_idx, stop_idx,
                                      tensor)
        return pred_values

    def __copy__(self) -> "SharedActorCriticLearner":
        """
        Creates a copy of the SharedActorCriticLearner instance.

        Returns:
            SharedActorCriticLearner: A copy of the current instance.
        """
        copy_ = SharedActorCriticLearner(self.input_dim, self.output_dim,
                                         self.tree_struct.copy(),
                                         self.optimizers[0].copy(),
                                         self.optimizers[1].copy(),
                                         self.params, self.verbose,
                                         self.device)
        copy_.iteration = self.iteration
        copy_.total_iterations = self.total_iterations
        if self._cpp_model is not None:
            copy_._cpp_model = GBRL_CPP(self._cpp_model)
        if self.student_model is not None:
            copy_.student_model = GBRL_CPP(self.student_model)
        return copy_


class SeparateActorCriticLearner(MultiGBTLearner):
    """
    Implements a separate actor-critic learner using two independent gradient
    boosted trees.

    This class extends MultiGBTLearner by maintaining two separate models:
    - One for policy learning (Actor).
    - One for value estimation (Critic).

    It provides separate `step_actor` and `step_critic` methods for updating
    the respective models.
    """
    def __init__(self, input_dim: int, output_dim: int, tree_struct: Dict,
                 policy_optimizer: Dict, value_optimizer: Dict,
                 params: Dict = dict(), verbose: int = 0, device: str = 'cpu'):
        """
        Initializes the SeparateActorCriticLearner with two independent GBT
        models.

        Args:
            input_dim (int): Number of input features.
            output_dim (int): Number of output dimensions.
            tree_struct (Dict): Dictionary containing tree structure parameters.
            policy_optimizer (Dict): Optimizer configuration for the policy (actor).
            value_optimizer (Dict): Optimizer configuration for the value function (critic).
            params (Dict, optional): Additional model parameters. Defaults to an empty dictionary.
            verbose (int, optional): Verbosity level for debugging. Defaults to 0.
            device (str, optional): Device to run the model on ('cpu' or 'cuda'). Defaults to 'cpu'.
        """
        if verbose > 0:
            print('****************************************')
            print(f'Separate GBRL Tree with input dim: {input_dim},'
                  f'output dim: {output_dim}, tree_struct: {tree_struct}'
                  f'policy_optimizer: {policy_optimizer}'
                  f'value_optimizer: {value_optimizer}')
            print('****************************************')
        super().__init__(input_dim, [output_dim - 1, 1], tree_struct,
                         [policy_optimizer, value_optimizer],
                         params, 2, verbose, device)

    def step(self, obs: NumericalData,  # type: ignore
             theta_grad: NumericalData, value_grad: NumericalData,
             model_idx: Optional[int] = None) -> None:
        """
        Performs a single gradient update step on both the policy and value
        models.

        Args:
            obs (NumericalData): Input observations.
            theta_grad (NumericalData): Gradient update for the policy (actor).
            value_grad (NumericalData): Gradient update for the value function (critic).
            model_idx (Optional[int], optional): Index of the model to update.
            If None, updates both models.
        """
        super().step(obs, [theta_grad, value_grad], model_idx=model_idx)

    def step_actor(self, obs: NumericalData,
                   theta_grad: NumericalData) -> None:
        """
        Performs a gradient update step for the policy (actor) model.

        Args:
            obs (NumericalData): Input observations.
            theta_grad (NumericalData): Gradient update for the policy (actor).
        """
        super().step(obs, theta_grad, model_idx=0)

    def step_critic(self, obs: NumericalData,
                    value_grad: NumericalData) -> None:
        """
        Performs a gradient update step for the value function (critic) model.

        Args:
            obs (NumericalData): Input observations.
            value_grad (NumericalData): Gradient update for the value function (critic).
        """
        super().step(obs, value_grad, model_idx=1)

    def distil(self, obs: NumericalData,  # type: ignore
               policy_targets: np.ndarray, value_targets: np.ndarray,
               params: Dict, verbose: int) -> Tuple[List[float], List[Dict]]:
        """
        Distills the trained model into a student model.

        Args:
            obs (NumericalData): Input observations.
            policy_targets (np.ndarray): Target values for the policy (actor).
            value_targets (np.ndarray): Target values for the value function (critic).
            params (Dict): Distillation parameters.
            verbose (int): Verbosity level.

        Returns:
            Tuple[List[float], List[Dict]]: The final loss values and updated parameters for distillation.
        """
        return super().distil(obs, [policy_targets, value_targets], params, verbose)  # type: ignore

    def predict(self, obs: NumericalData,  # type: ignore
                requires_grad: bool = True, start_idx: int = 0,
                stop_idx: Optional[int] = None, tensor: bool = True) -> \
            List[NumericalData]:
        """
        Predicts both the policy and value outputs for the given observations.

        Args:
            obs (NumericalData): Input observations.
            requires_grad (bool, optional): Whether to compute gradients. Defaults to True.
            start_idx (int, optional): Start index for prediction. Defaults to 0.
            stop_idx (int, optional): Stop index for prediction. Defaults to None.
            tensor (bool, optional): Whether to return a tensor. Defaults to True.

        Returns:
            List[NumericalData]: Predicted policy outputs and value function outputs.
        """
        preds = super().predict(obs, requires_grad, start_idx, stop_idx, tensor)
        assert isinstance(preds, list), "SeparateActorCriticLearner.predict should always return a list"
        return preds

    def predict_policy(self, obs: NumericalData,
                       requires_grad: bool = True, start_idx: int = 0,
                       stop_idx: Optional[int] = None, tensor: bool = True) -> NumericalData:
        """
        Predicts the policy (actor) output for the given observations.

        Args:
            obs (NumericalData): Input observations.
            requires_grad (bool, optional): Whether to compute gradients. Defaults to True.
            start_idx (int, optional): Start index for prediction. Defaults to 0.
            stop_idx (int, optional): Stop index for prediction. Defaults to None.
            tensor (bool, optional): Whether to return a tensor. Defaults to True.

        Returns:
            NumericalData: Predicted policy outputs.
        """
        preds = super().predict(obs, requires_grad, start_idx, stop_idx, tensor, model_idx=0)
        assert not isinstance(preds, list), "SeparateActorCriticLearner.predict_policy should never return a list"
        return preds

    def predict_critic(self, obs: NumericalData,
                       requires_grad: bool = True, start_idx: int = 0,
                       stop_idx: Optional[int] = None, tensor: bool = True) -> NumericalData:
        """
        Predicts the value function (critic) output for the given observations.

        Args:
            obs (NumericalData): Input observations.
            requires_grad (bool, optional): Whether to compute gradients. Defaults to True.
            start_idx (int, optional): Start index for prediction. Defaults to 0.
            stop_idx (int, optional): Stop index for prediction. Defaults to None.
            tensor (bool, optional): Whether to return a tensor. Defaults to True.

        Returns:
            NumericalData: Predicted value function outputs.
        """
        preds = super().predict(obs, requires_grad, start_idx, stop_idx, tensor, model_idx=1)
        assert not isinstance(preds, list), "SeparateActorCriticLearner.predict_critic should never return a list"
        return preds

    def __copy__(self) -> "SeparateActorCriticLearner":
        """
        Creates a copy of the SeparateActorCriticLearner instance.

        Returns:
            SeparateActorCriticLearner: A new instance with the same parameters and structure.
        """
        opts = [opt.copy() if opt is not None else opt
                for opt in self.optimizers
                ]
        copy_ = SeparateActorCriticLearner(self.input_dim, self.output_dim,
                                           self.tree_struct.copy(),
                                           opts[0], opts[1],
                                           self.params,
                                           self.verbose,
                                           self.device)
        assert copy_._cpp_models is not None, "Copy models not initialized"
        assert copy_.student_models is not None, "Copy student models not initialized"
        copy_.iteration = self.iteration
        copy_.total_iterations = self.total_iterations
        if self._cpp_models is not None:
            for i in range(self.n_learners):
                copy_._cpp_models[i] = GBRL_CPP(self._cpp_models[i])
                if self.student_models is not None:
                    copy_.student_models[i] = GBRL_CPP(self.student_models[i])
        return copy_



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
from typing import Dict, List, Optional, Union

import numpy as np
import torch as th

from gbrl.learners.gbt_learner import GBTLearner
from gbrl.models.base import BaseGBT
from gbrl.common.utils import (NumericalData, clip_grad_norm, setup_optimizer,
                               validate_array)


class GBTModel(BaseGBT):
    """
    General class for gradient boosting trees

    """
    def __init__(self,
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int,
                 optimizers: Union[Dict, List[Dict]],
                 params: Dict = dict(),
                 verbose: int = 0,
                 device: str = 'cpu'):
        """
        Initializes the GBT model.

        Args:
            tree_struct (Dict): Dictionary containing
            tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
            output_dim (int): output dimension.
            optimizers (Union[Dict, List[Dict]]): dictionary containing
            optimizer parameters or a list of dictionaries containing
            optimizer parameters.
                SGD optimizer must contain as:
                    'algo': 'SGD'
                    'lr' Union[float, str]: learning rate value. Constant learning rate is used by default
                Adam optimizer must contain (CPU only):
                    'algo' (str): 'Adam'
                    'lr' Union[float, str]: learning rate value. Constant learning rate is used by default
                    'beta_1' (float): 0.99
                    'beta_2' (float): 0.999
                    'eps' (float): 1.0e-8
                All optimizers must contain:
                    'start_idx'
                    'stop_idx'
                Setting scheduler type:
                Available schedulers are Constant and Linear. Constant is default, Linear is CPU only.
                To specify a linear scheduler, 3 additional arguments must be
                added to an optimizer dict.
                    'init_lr' (str): "lin_<value>"
                    'stop_lr' (float): minimum lr value
                    'T' (int): number of total expected boosting trees,
                               used to calculate the linear scheduling internally.
                               Can be manually calculated per algorithm
                               according to the total number of training steps.
             params (Dict, optional): GBRL parameters such as:
                control_variates (bool): use control variates (variance
                reduction technique CPU only).
                split_score_func (str): "cosine" or "l2"
                generator_type - (str): candidate generation method "Quantile" or "Uniform".
                feature_weights - (list[float]): Per-feature multiplication
                weights used when choosing the best split. Weights should be >= 0
            verbose (int, optional): verbosity level. Defaults to 0.
            device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """
        super().__init__()
        if optimizers is not None:
            if isinstance(optimizers, dict):
                optimizers = [optimizers]
            optimizers = [setup_optimizer(opt) for opt in optimizers]

        self.learner = GBTLearner(input_dim, output_dim,
                                  tree_struct, optimizers,
                                  params, verbose, device)
        self.learner.reset()
        self.grad = None
        self.input = None
        self.params = None

    def set_bias(self, bias: NumericalData) -> None:
        """Sets GBRL bias

        Args:
            bias (NumericalData)
        """
        assert self.learner is not None, "Learner model not yet initialized"
        if isinstance(bias, th.Tensor):
            bias = bias.clone().detach().cpu().numpy()
        # GBRL works with 2D numpy arrays.
        if len(bias.shape) == 1:
            bias = bias[:, np.newaxis]
        self.learner.set_bias(bias.astype(np.single))   # type: ignore

    def set_bias_from_targets(self, targets: Union[np.ndarray,
                                                   th.Tensor]) -> None:
        """Sets bias as mean of targets

        Args:
            targets (NumericalData): Targets
        """
        assert self.learner is not None, "Learner model not yet initialized"
        if isinstance(targets, th.Tensor):
            arr = targets.clone().detach().cpu().numpy()
        else:
            arr = targets.copy()
        # GBRL works with 2D numpy arrays.
        if len(arr.shape) == 1:
            arr = arr[:, np.newaxis]
        mean_arr = np.mean(arr, axis=0)
        if isinstance(mean_arr, float):
            mean_arr = np.ndarray([mean_arr])  # type: ignore
        # GBRL only works with float32
        self.learner.set_bias(mean_arr)   # type: ignore

    def step(self,  X: Optional[NumericalData] = None,
             grad: Optional[NumericalData] = None,
             max_grad_norm: Optional[float] = None) -> None:
        """
        Perform a boosting step (fits a single tree on the gradients)

        Args:
            X (NumericalData): inputs
            max_grad_norm (float, optional): perform gradient clipping by norm. Defaults to None.
            grad (Optional[NumericalData], optional): manually calculated gradients. Defaults to None.
        """
        assert self.learner is not None, "Learner model not yet initialized"
        if X is None:
            assert self.input is not None, ("Cannot update trees without input."
                                            "Make sure model is called with requires_grad=True")
            X = self.input
        n_samples = len(X)
        grad = grad if grad is not None else self.params.grad.detach() * n_samples  # type: ignore

        grad = clip_grad_norm(grad, max_grad_norm)
        validate_array(grad)
        self.learner.step(X, grad)
        self.grad = grad
        self.input = None

    def fit(self, X: NumericalData,
            targets: NumericalData,
            iterations: int, shuffle: bool = True,
            loss_type: str = 'MultiRMSE') -> float:
        """
        Fit multiple iterations (as in supervised learning)

        Args:
            X (NumericalData): inputs
            targets (NumericalData): targets
            iterations (int): number of boosting iterations
            shuffle (bool, optional): Shuffle dataset. Defaults to True.
            loss_type (str, optional): Loss to use (only MultiRMSE is currently implemented ). Defaults to 'MultiRMSE'.

        Returns:
            float: final loss over all examples.
        """
        assert self.learner is not None, "Learner model not yet initialized"
        return self.learner.fit(X, targets, iterations, shuffle, loss_type)

    @classmethod
    def load_learner(cls, load_name: str, device: str) -> "GBTModel":
        """Loads GBRL model from a file

        Args:
            load_name (str): full path to file name

        Returns:
            GBRL instance
        """
        instance = cls.__new__(cls)
        instance.learner = GBTLearner.load(load_name, device)
        instance.params = None
        return instance

    def __call__(self, X: Union[th.Tensor, np.ndarray],
                 requires_grad: bool = True, start_idx: int = 0,
                 stop_idx: Optional[int] = None, tensor: bool = True) -> Union[th.Tensor, np.ndarray]:
        """
        Returns GBRL's output as either a Tensor or a numpy array. if
        `requires_grad=True` then stores differentiable parameters
        in self.params. Return type/device is identical to the input type/device.

        Args:
            X (Union[th.Tensor, np.ndarray]): Input
            requires_grad (bool, optional). Defaults to True.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction
            (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Union[th.Tensor, np.ndarray]: Returns model predictions
        """
        assert self.learner is not None, "Learner model not yet initialized"
        y_pred = self.learner.predict(X, requires_grad, start_idx, stop_idx, tensor)
        if requires_grad:
            self.grad = None
            self.params = y_pred
            self.input = X
        return y_pred

    def print_tree(self, tree_idx: int) -> None:
        """Prints tree information

        Args:
            tree_idx (int): tree index to print
        """
        assert self.learner is not None, "Learner model not yet initialized"
        self.learner.print_tree(tree_idx)

    def plot_tree(self, tree_idx: int, filename: str) -> None:
        """Plots tree using (only works if GBRL was compiled with graphviz)

        Args:
            tree_idx (int): tree index to plot
            filename (str): .png filename to save
        """
        assert self.learner is not None, "Learner model not yet initialized"
        self.learner.plot_tree(tree_idx, filename)

    def copy(self) -> "GBTModel":
        """Copy class instance

        Returns:
            GradientBoostingTrees: copy of current instance. The actual type
            will be the type of the subclass that calls this method.
        """
        return self.__copy__()

    def __copy__(self) -> "GBTModel":
        assert self.learner is not None, "Learner model not yet initialized"

        learner = self.learner.copy()
        copy_ = GBTModel(learner.tree_struct, learner.input_dim,
                         learner.output_dim, learner.optimizers,  # type: ignore
                         learner.params, learner.verbose, learner.device)
        copy_.learner = learner  # type: ignore
        return copy_



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
from typing import Dict, Optional, Tuple, cast

import numpy as np
from torch import Tensor

from gbrl.common.utils import (NumericalData, clip_grad_norm,
                               concatenate_arrays, constant_like,
                               ensure_leaf_tensor_or_array, numerical_dtype,
                               setup_optimizer, validate_array)
from gbrl.learners.gbt_learner import GBTLearner
from gbrl.models.base import BaseGBT


class ParametricActor(BaseGBT):
    """
    GBRL model for a ParametricActor ensemble. ParametricActor outputs
    a single parameter per action dimension, allowing deterministic or
    stochastic behavior
    (e.g., for discrete action spaces).
    """
    def __init__(self,
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int,
                 policy_optimizer: Dict,
                 params: Dict = dict(),
                 bias: Optional[np.ndarray] = None,
                 verbose: int = 0,
                 device: str = 'cpu'):
        """
        Initializes the ParametricActor model.

        Args:
            tree_struct (Dict): Dictionary containing tree structure
            information:
                - max_depth (int): Maximum tree depth.
                - grow_policy (str): 'greedy' or 'oblivious'.
                - n_bins (int): Number of bins per feature for candidate
                generation.
                - min_data_in_leaf (int): Minimum number of samples in a leaf.
                - par_th (int): Minimum number of samples for parallelizing on
                CPU.
            input_dim (int): Input feature dimension.
            output_dim (int): Output feature dimension.
            policy_optimizer (Dict): Dictionary containing policy optimizer
            parameters.
            params (Dict, optional): Additional GBRL parameters.
            bias (np.ndarray, optional): Bias initialization, defaults to zero.
            verbose (int, optional): Verbosity level. Defaults to 0.
            device (str, optional): Compute device ('cpu' or 'cuda'). Defaults
            to 'cpu'.
        """
        policy_optimizer = setup_optimizer(policy_optimizer, prefix='policy_')
        super().__init__()
        bias = bias if bias is not None else np.zeros(output_dim,
                                                      dtype=numerical_dtype)
        # init model
        self.learner = GBTLearner(input_dim, output_dim, tree_struct, policy_optimizer,
                                  params, verbose, device)
        self.learner.reset()
        self.learner.set_bias(bias)

    def step(self, observations: Optional[NumericalData] = None,
             policy_grad: Optional[NumericalData] = None,
             policy_grad_clip: Optional[float] = None) -> None:
        """
        Performs a single boosting iteration.

        Args:
            observations (NumericalData):
            policy_grad_clip (float, optional): . Defaults to None.
            policy_grad (Optional[NumericalData], optional): manually
            calculated gradients. Defaults to None.
        """
        assert self.learner is not None, "Learner model not yet initialized"
        assert policy_grad is not None or (self.params is not None and isinstance(self.params, Tensor)), \
            "Cannot update trees without gradients."
        if observations is None:
            assert self.input is not None, "Cannot update trees without input."
            "Make sure model is called with requires_grad=True"
            observations = self.input
        n_samples = len(observations)

        policy_grad = policy_grad if policy_grad is not None else self.params.grad.detach() * n_samples  # type: ignore
        policy_grad = clip_grad_norm(policy_grad, policy_grad_clip)
        validate_array(policy_grad)

        self.learner.step(observations, policy_grad)
        self.grad = policy_grad
        self.input = None

    def __call__(self, observations: NumericalData,
                 requires_grad: bool = True, start_idx: int = 0,
                 stop_idx: Optional[int] = None, tensor: bool = True) -> NumericalData:
        """
        Returns actor output as Tensor. If `requires_grad=True`, stores
        differentiable parameters in `self.params`.

        Args:
            observations (NumericalData): Input observations.
            requires_grad (bool, optional): Whether to compute gradients.
            Defaults to True.
            start_idx (int, optional): Start tree index for prediction.
            Defaults to 0.
            stop_idx (Optional[int], optional): Stop tree index for prediction.
            Defaults to None.
            tensor (bool, optional): Whether to return a PyTorch Tensor.
            Defaults to True.

        Returns:
            NumericalData: GBRL outputs - a single parameter per action
            dimension.
        """
        assert self.learner is not None, "Learner model not yet initialized"
        params = self.learner.predict(observations, requires_grad, start_idx,
                                      stop_idx, tensor)
        if requires_grad:
            self.grads = None
            self.params = params
            self.input = observations
        return params

    def __copy__(self) -> "ParametricActor":
        """
        Creates a copy of the ParametricActor instance.

        Returns:
            ParametricActor: A copy of the current model.
        """
        assert self.learner is not None, "Learner model not yet initialized"
        learner = cast(GBTLearner, self.learner.copy())
        copy_ = ParametricActor(learner.tree_struct, learner.input_dim,
                                learner.output_dim, learner.optimizers[0],
                                learner.params, learner.get_bias(),
                                learner.verbose, learner.device)
        copy_.learner = learner
        return copy_


class GaussianActor(BaseGBT):
    """
    GBRL model for an actor ensemble used in algorithms such as SAC.
    This model outputs the mean (`mu`) and log standard deviation (`log_std`)
    of a Gaussian distribution, allowing stochastic action selection.
    """
    def __init__(self,
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int,
                 mu_optimizer: Dict,
                 std_optimizer: Optional[Dict] = None,
                 log_std_init: float = -2,
                 params: Dict = dict(),
                 bias: Optional[np.ndarray] = None,
                 verbose: int = 0,
                 device: str = 'cpu'):
        """
        Initializes the GaussianActor model.

        Args:
         tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate
                generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on
                CPU.
        output_dim (int): output dimension.
        mu_optimizer Dict: dictionary containing Gaussian mean optimizer
        parameters. (see GradientBoostingTrees for optimizer details)
        std_optimizer Dict: dictionary containing Gaussian sigma optimizer
        parameters. (see GradientBoostingTrees for optimizer details)
        log_std_init (float): initial value of log_std
        params (Dict, optional): GBRL parameters such as:
            control_variates (bool): use control variates (variance reduction
            technique CPU only).
            split_score_func (str): "cosine" or "l2"
            generator_type- (str): candidate generation method "Quantile" or
            "Uniform".
            feature_weights - (list[float]): Per-feature multiplication
            weights used when choosing the best split. Weights should be >= 0
        bias (np.ndarray, optional): manually set a bias. Defaults to None =
        np.zeros.
        verbose (int, optional): verbosity level. Defaults to 0.
        device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to
        'cpu'.
        """
        super().__init__()
        mu_optimizer = setup_optimizer(mu_optimizer, prefix='mu_')

        bias = bias if bias is not None else np.zeros(output_dim,
                                                      dtype=numerical_dtype)
        policy_dim = output_dim
        if std_optimizer is not None:
            std_optimizer = setup_optimizer(std_optimizer, prefix='std_')
            policy_dim = output_dim // 2
            bias[policy_dim:] = log_std_init*np.ones(policy_dim,
                                                     dtype=numerical_dtype)
        self.log_std_init = log_std_init
        self.fixed_std = std_optimizer is None
        self.policy_dim = policy_dim

        # init model
        self.learner = GBTLearner(input_dim, output_dim, tree_struct,
                                  [mu_optimizer, std_optimizer], params,
                                  verbose, device)
        self.learner.reset()
        self.learner.set_bias(bias)

    def step(self, observations: Optional[NumericalData] = None,
             mu_grad: Optional[NumericalData] = None,
             log_std_grad: Optional[NumericalData] = None,
             mu_grad_clip: Optional[float] = None,
             log_std_grad_clip: Optional[float] = None) -> None:
        """
        Performs a single boosting iteration.

        Args:
            observations (NumericalData): Input observations.
            mu_grad (Optional[NumericalData], optional):
            Manually computed mean gradients.
            log_std_grad (Optional[NumericalData], optional):
            Manually computed log standard deviation gradients.
            mu_grad_clip (Optional[float], optional): Gradient clipping for
            mean. Defaults to None.
            log_std_grad_clip (Optional[float], optional): Gradient clipping
            for log standard deviation. Defaults to None.
        """
        assert self.learner is not None, "Learner model not yet initialized"
        if observations is None:
            assert self.input is not None, "Cannot update trees without input."
            "Make sure model is called with requires_grad=True"
            observations = self.input
        n_samples = len(observations)
        mu_grad = mu_grad if mu_grad is not None else \
            self.params[0].grad.detach() * n_samples  # type: ignore
        mu_grad = clip_grad_norm(mu_grad, mu_grad_clip)  # type: ignore

        if not self.fixed_std:
            log_std_grad = log_std_grad if log_std_grad is not None else \
                self.params[1].grad.detach() * n_samples  # type: ignore
            log_std_grad = clip_grad_norm(log_std_grad, log_std_grad_clip)  # type: ignore
            theta_grad = concatenate_arrays(mu_grad, log_std_grad)
        else:
            theta_grad = mu_grad

        validate_array(theta_grad)

        self.learner.step(observations, theta_grad)
        self.grad = mu_grad
        if not self.fixed_std:
            self.grad = (mu_grad, log_std_grad)  # type: ignore
        self.input = None

    def __call__(self, observations: NumericalData,
                 requires_grad: bool = True,
                 start_idx: int = 0, stop_idx: Optional[int] = None,
                 tensor: bool = True) -> Tuple[NumericalData,
                                               NumericalData]:
        """
        Returns actor's outputs as tensor. If `requires_grad=True` then stores
           differentiable parameters in self.params. Return type/device is
           identical to the input type/device.
           Requires_grad is ignored if input is a numpy array.
        Args:
            observations (NumericalData)
            requires_grad bool: Defaults to None.
            start_idx (int, optional): start tree index for prediction.
            Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses
            all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a
            numpy array. Defaults to True.

        Returns:
            NumericalData: Gaussian parameters
        """
        assert self.learner is not None, "Learner model not yet initialized"
        theta = self.learner.predict(observations, requires_grad, start_idx,
                                     stop_idx, tensor)
        mean_actions = theta if self.fixed_std else theta[:, :self.policy_dim]
        if not self.fixed_std:
            mean_actions = ensure_leaf_tensor_or_array(mean_actions, tensor=True, requires_grad=requires_grad,
                                                       device=self.learner.device)
        log_std = constant_like(theta, self.log_std_init) if self.fixed_std else theta[:, self.policy_dim:]
        log_std = ensure_leaf_tensor_or_array(log_std, tensor=True, requires_grad=False if
                                              self.fixed_std else
                                              requires_grad,
                                              device=self.learner.device)
        if requires_grad:
            self.grad = None
            self.params = mean_actions, log_std  # type: ignore
            self.input = observations
        return mean_actions, log_std  # type: ignore

    def __copy__(self) -> "GaussianActor":
        """
        Creates a copy of the GaussianActor instance.

        Returns:
            GaussianActor: A copy of the current model.
        """
        assert self.learner is not None, "Learner model not yet initialized"
        learner = cast(GBTLearner, self.learner.copy())
        copy_ = GaussianActor(learner.tree_struct, learner.input_dim,
                              learner.output_dim, learner.optimizers[0],
                              learner.optimizers[1],  self.log_std_init,
                              learner.params,  # type: ignore
                              learner.get_bias(), learner.verbose,  # type: ignore
                              learner.device)  # type: ignore
        copy_.learner = learner
        return copy_



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
import os
import shutil
import tempfile
import unittest
from typing import Union, Tuple

import numpy as np
import torch as th
from sklearn.tree import DecisionTreeRegressor
import shap
from torch.nn.functional import mse_loss

from sklearn import datasets

from gbrl import cuda_available
from gbrl.models.gbt import GBTModel
from gbrl.models.actor_critic import ActorCritic


N_EPOCHS = 100


def rmse(preds: Union[np.array, th.Tensor],
         targets: Union[np.array, th.Tensor]) -> Tuple[float, np.array]:
    if isinstance(preds, th.Tensor):
        preds = preds.detach().cpu().numpy()
    if isinstance(targets, th.Tensor):
        targets = targets.detach().cpu().numpy()

    preds = preds.squeeze()
    targets = targets.squeeze()

    loss = np.sqrt(np.mean((preds - targets)**2))
    return loss, (preds - targets)


def rmse_model(model, X, y, n_epochs, device='cpu'):
    y_ = th.tensor(y, dtype=th.float32, device=device)
    X_ = X.copy()
    epoch = 0
    while epoch < n_epochs:
        y_pred = model(X_, requires_grad=True)
        loss = 0.5*mse_loss(y_pred, y_) * y_.shape[1]
        loss.backward()
        model.step()
        print(f"epoch: {epoch} loss: {(loss / y_.shape[1]).sqrt()}")
        epoch += 1
    y_pred = model(X_)
    loss = (0.5*mse_loss(y_pred, y_)).sqrt().item()
    return loss


def ac_rmse_model(model, X, y, n_epochs, device='cpu'):
    y_ac = th.tensor(y[:, :-1], dtype=th.float32, device=device)
    y_value = th.tensor(y[:, -1], dtype=th.float32, device=device)
    X_ = X.copy()
    epoch = 0
    while epoch < n_epochs:
        theta, value = model(X_, requires_grad=True)
        loss_theta = 0.5*mse_loss(theta, y_ac) * y_ac.shape[1]
        loss_theta.backward()
        loss_value = 0.5*mse_loss(value, y_value)
        loss_value.backward()
        model.step()
        print(f"epoch: {epoch} loss_theta: {loss_theta.sqrt():.5f} "
              f"loss_value: {(loss_value).sqrt():.5f}")
        epoch += 1
    theta, value = model(X_)
    loss_theta = (0.5*mse_loss(theta, y_ac)).sqrt().item()
    loss_value = (0.5*mse_loss(value, y_value)).sqrt().item()
    return loss_theta, loss_value


class TestGBTMulti(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print('Loading data...')
        # Imagine this loads your actual data
        try:
            X, y = datasets.load_diabetes(return_X_y=True, as_frame=False,
                                          scaled=False)
        except TypeError:  # python3.7 uses older version of scikit-learn
            X, y = datasets.load_diabetes(return_X_y=True, as_frame=False)
        out_dim = 1 if len(y.shape) == 1 else y.shape[1]
        if out_dim == 1:
            y = y[:, np.newaxis]
        y_fake = y.copy()
        n_cols = 10
        for _ in range(n_cols - 1):
            y = np.concatenate([y, y_fake], axis=1)
        out_dim = y.shape[1]
        cls.data = (X, y)
        cls.out_dim = out_dim
        cls.input_dim = X.shape[1]
        cls.n_epochs = 100
        cls.test_dir = tempfile.mkdtemp()
        cls.tree_struct = {'max_depth': 4,
                           'n_bins': 256, 'min_data_in_leaf': 0,
                           'par_th': 2,
                           'grow_policy': 'greedy'}
        cls.sgd_optimizer = {'algo': 'SGD',
                             'lr': 1.0,
                             'start_idx': 0,
                             'stop_idx': out_dim
                             }

    @classmethod
    def tearDownClass(cls):
        # Remove the directory after the test
        shutil.rmtree(cls.test_dir)

    def test_cosine_cpu(self):
        print("Running Multi test_cosine_cpu")
        X, y = self.data
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        value = 2.0
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir, 'test_cosine_cpu'))

    def test_shap_cpu(self):
        print("Running test_shap_cpu")
        X, y = self.data
        tree_struct = {'max_depth': 3,
                       'n_bins': 256, 'min_data_in_leaf': 1,
                       'par_th': 2,
                       'grow_policy': 'greedy'}

        params = dict({"control_variates": False, "split_score_func": "L2",
                       "generator_type": "Uniform"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.learner.step(X, y)
        gbrl_shap = model.tree_shap(0, X[0, :])
        clf = DecisionTreeRegressor(max_depth=3).fit(X, y)
        target_shap = shap.TreeExplainer(clf).shap_values(X[0])
        self.assertTrue(np.allclose(gbrl_shap, target_shap, rtol=1e-3),
                        'GBRL sHAP values are not close to target '
                        'SHAP values')

    def test_cosine_adam_cpu(self):
        print("Running Multi test_cosine_adam_cpu")
        X, y = self.data
        optimizer = {'algo': 'Adam',
                     'lr': 1.0,
                     'start_idx': 0,
                     'stop_idx': self.out_dim
                     }
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        value = 50.0
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir, 'test_cosine_adam_cpu'))

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_cosine_gpu(self):
        print("Running Multi test_cosine_gpu")
        X, y = self.data
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs, device='cuda')
        value = 2.0
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir, 'test_cosine_gpu'))

    def test_cosine_oblivious_cpu(self):
        print("Running Multi test_cosine_oblivious_cpu")
        X, y = self.data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        self.assertTrue(loss < 12, f'Expected loss = {loss} < 12')
        model.save_learner(os.path.join(self.test_dir,
                                        'test_cosine_oblivious_cpu'))

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_cosine_oblivious_gpu(self):
        print("Running Multi test_cosine_oblivious_gpu")
        X, y = self.data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False,
                       "split_score_func": "Cosine"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs, device='cuda')
        value = 12
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir,
                                        'test_cosine_oblivious_gpu'))

    def test_l2_cpu(self):
        print("Running Multi test_l2_cpu")
        X, y = self.data
        params = dict({"control_variates": False, "split_score_func": "L2"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        value = 0.5
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir, 'test_l2_cpu'))

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_l2_gpu(self):
        print("Running Multi test_l2_gpu")
        X, y = self.data
        params = dict({"control_variates": False, "split_score_func": "L2"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=self.tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs, device='cuda')
        value = 0.5
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir, 'test_l2_gpu'))

    def test_l2_oblivious_cpu(self):
        print("Running Multi test_l2_oblivious_cpu")
        X, y = self.data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False, "split_score_func": "L2"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cpu')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs)
        value = 10.0
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir,
                                        'test_l2_oblivious_cpu'))

    def test_1shared_cpu(self):
        print("Running shared_cpu")
        X, y = self.data
        policy_optimizer = {
            'policy_algo': 'SGD',
            'policy_lr': 1.0,
            'start_idx': 0,
            'stop_idx': self.out_dim - 1
        }
        value_optimizer = {
                    'value_algo': 'SGD',
                    'value_lr': 0.1,
                    'start_idx': self.out_dim - 1,
                    'stop_idx': self.out_dim
                }
        params = dict({"control_variates": False,
                       "split_score_func": "cosine"})
        model = ActorCritic(
                            input_dim=self.input_dim,
                            output_dim=self.out_dim,
                            tree_struct=self.tree_struct,
                            policy_optimizer=policy_optimizer,
                            shared_tree_struct=True,
                            value_optimizer=value_optimizer,
                            params=params,
                            verbose=0,
                            device='cpu')
        policy_loss, value_loss = ac_rmse_model(model, X, y, self.n_epochs)
        policy_value = 10.0
        value_value = 30
        self.assertTrue(policy_loss < policy_value,
                        f'Expected loss = {policy_loss} < {policy_value}')
        self.assertTrue(value_loss < value_value,
                        f'Expected loss = {value_loss} < {value_value}')
        model.save_learner(os.path.join(self.test_dir, 'test_shared_cpu'))

    def test_1separate_cpu(self):
        print("Running separate_cpu")
        X, y = self.data
        policy_optimizer = {
            'policy_algo': 'SGD',
            'policy_lr': 1.0,
            'start_idx': 0,
            'stop_idx': self.out_dim - 1
        }
        value_optimizer = {
            'value_algo': 'SGD',
            'value_lr': 0.1,
            'start_idx': 0,
            'stop_idx': 1
        }
        params = dict({"control_variates": False,
                       "split_score_func": "cosine"})
        model = ActorCritic(input_dim=self.input_dim,
                            output_dim=self.out_dim,
                            tree_struct=self.tree_struct,
                            policy_optimizer=policy_optimizer,
                            shared_tree_struct=False,
                            value_optimizer=value_optimizer,
                            params=params,
                            verbose=0,
                            device='cpu')
        policy_loss, value_loss = ac_rmse_model(model, X, y, self.n_epochs)
        policy_value = 3.0
        value_value = 30
        self.assertTrue(policy_loss < policy_value,
                        f'Expected loss = {policy_loss} < {policy_value}')
        self.assertTrue(value_loss < value_value,
                        f'Expected loss = {value_loss} < {value_value}')
        model.save_learner(os.path.join(self.test_dir, 'test_separate_cpu'))

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_1separate_gpu(self):
        print("Running separate_gpu")
        X, y = self.data
        policy_optimizer = {
            'policy_algo': 'SGD',
            'policy_lr': 1.0,
            'start_idx': 0,
            'stop_idx': self.out_dim - 1
        }
        value_optimizer = {
                    'value_algo': 'SGD',
                    'value_lr': 0.1,
                    'start_idx': 0,
                    'stop_idx': 1
                }
        params = dict({"control_variates": False,
                       "split_score_func": "cosine"})
        model = ActorCritic(
                            input_dim=self.input_dim,
                            output_dim=self.out_dim,
                            tree_struct=self.tree_struct,
                            policy_optimizer=policy_optimizer,
                            shared_tree_struct=False,
                            value_optimizer=value_optimizer,
                            params=params,
                            verbose=0,
                            device='cuda')
        policy_loss, value_loss = ac_rmse_model(model, X, y, self.n_epochs,
                                                device='cuda')
        policy_value = 2.0
        value_value = 30
        self.assertTrue(policy_loss < policy_value,
                        f'Expected loss = {policy_loss} < {policy_value}')
        self.assertTrue(value_loss < value_value,
                        f'Expected loss = {value_loss} < {value_value}')
        model.save_learner(os.path.join(self.test_dir, 'test_separate_gpu'))

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_1shared_gpu(self):
        print("Running shared_gpu")
        X, y = self.data
        policy_optimizer = {
            'policy_algo': 'SGD',
            'policy_lr': 1.0,
            'start_idx': 0,
            'stop_idx': self.out_dim - 1
        }
        value_optimizer = {
                    'value_algo': 'SGD',
                    'value_lr': 0.1,
                    'start_idx': self.out_dim - 1,
                    'stop_idx': self.out_dim
                }
        params = dict({"control_variates": False,
                       "split_score_func": "cosine"})
        model = ActorCritic(input_dim=self.input_dim,
                            output_dim=self.out_dim,
                            tree_struct=self.tree_struct,
                            policy_optimizer=policy_optimizer,
                            shared_tree_struct=True,
                            value_optimizer=value_optimizer,
                            params=params,
                            verbose=0,
                            device='cuda')
        policy_loss, value_loss = ac_rmse_model(model, X, y, self.n_epochs,
                                                device='cuda')
        policy_value = 10.0
        value_value = 30
        self.assertTrue(policy_loss < policy_value,
                        f'Expected loss = {policy_loss} < {policy_value}')
        self.assertTrue(value_loss < value_value,
                        f'Expected loss = {value_loss} < {value_value}')
        model.save_learner(os.path.join(self.test_dir, 'test_shared_gpu'))

    @unittest.skipIf(not cuda_available(),
                     "cuda not available skipping over gpu tests")
    def test_l2_oblivious_gpu(self):
        print("Running Multi test_l2_oblivious_gpu")
        X, y = self.data
        tree_struct = {'max_depth': 4,
                       'n_bins': 256, 'min_data_in_leaf': 0,
                       'par_th': 2,
                       'grow_policy': 'oblivious'}
        params = dict({"control_variates": False, "split_score_func": "L2"})
        model = GBTModel(input_dim=self.input_dim,
                         output_dim=self.out_dim,
                         tree_struct=tree_struct,
                         optimizers=self.sgd_optimizer,
                         params=params,
                         verbose=0,
                         device='cuda')
        model.set_bias_from_targets(y)
        loss = rmse_model(model, X, y, self.n_epochs, device='cuda')
        value = 10.0
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')
        model.save_learner(os.path.join(self.test_dir,
                                        'test_l2_oblivious_gpu'))

    def test_loading(self):
        X, y = self.data

        model = GBTModel.load_learner(os.path.join(self.test_dir,
                                                   'test_cosine_cpu'),
                                      device='cpu')
        y_pred = model(X, requires_grad=False, tensor=False)
        loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
        self.assertTrue(loss < 2.0, f'Expected loss = {loss} < 2.0')

        model = GBTModel.load_learner(os.path.join(self.test_dir, 'test_l2_cpu'),
                                      device='cpu')
        y_pred = model(X, requires_grad=False, tensor=False)
        loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
        self.assertTrue(loss < 0.5, f'Expected loss = {loss} < 0.5')

        model = ActorCritic.load_learner(os.path.join(self.test_dir,
                                                      'test_shared_cpu'),
                                         device='cpu')
        policy_y, value_y = model(X, requires_grad=False, tensor=False)
        policy_loss, _ = rmse(policy_y, y[:, :-1])
        value_loss, _ = rmse(value_y, y[:, -1])
        policy_value = 10.0
        value_value = 30
        self.assertTrue(policy_loss < policy_value,
                        f'Expected loss = {policy_loss} < {policy_value}')
        self.assertTrue(value_loss < value_value,
                        f'Expected loss = {value_loss} < {value_value}')

        model = ActorCritic.load_learner(
            os.path.join(self.test_dir, 'test_separate_cpu'), device='cpu')
        policy_y, value_y = model(X, requires_grad=False, tensor=False)
        policy_loss, _ = rmse(policy_y, y[:, :-1])
        value_loss, _ = rmse(value_y, y[:, -1])
        policy_value = 2.0
        value_value = 30
        self.assertTrue(policy_loss < policy_value,
                        f'Expected loss = {policy_loss} < {policy_value}')
        self.assertTrue(value_loss < value_value,
                        f'Expected loss = {value_loss} < {value_value}')

        if (cuda_available()):
            model = GBTModel.load_learner(
                os.path.join(self.test_dir, 'test_cosine_gpu'), device='cuda')
            y_pred = model(X, requires_grad=False, tensor=False)
            loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
            self.assertTrue(loss < 2.0, f'Expected loss = {loss} < 2.0')

            model = GBTModel.load_learner(
                os.path.join(self.test_dir, 'test_cosine_oblivious_gpu'),
                device='cuda')
            y_pred = model(X, requires_grad=False, tensor=False)
            loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
            self.assertTrue(loss < 12.0, f'Expected loss = {loss} < 12.0')

            model = GBTModel.load_learner(os.path.join(self.test_dir,
                                                       'test_l2_gpu'),
                                          device='cuda')
            y_pred = model(X, requires_grad=False, tensor=False)
            loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
            self.assertTrue(loss < 0.5, f'Expected loss = {loss} < 0.5')

            model = GBTModel.load_learner(
                os.path.join(self.test_dir, 'test_l2_oblivious_gpu'),
                device='cuda')
            y_pred = model(X, requires_grad=False, tensor=False)
            loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
            self.assertTrue(loss < 10.0, f'Expected loss = {loss} < 10.0')

            model = ActorCritic.load_learner(
                os.path.join(self.test_dir, 'test_shared_gpu'), device='cuda')
            policy_y, value_y = model(X, requires_grad=False, tensor=False)
            policy_loss, _ = rmse(policy_y, y[:, :-1])
            value_loss, _ = rmse(value_y, y[:, -1])
            policy_value = 10.0
            value_value = 30
            self.assertTrue(policy_loss < policy_value,
                            f'Expected loss = {policy_loss} < {policy_value}')
            self.assertTrue(value_loss < value_value,
                            f'Expected loss = {value_loss} < {value_value}')

            model = ActorCritic.load_learner(
                os.path.join(self.test_dir, 'test_separate_gpu'),
                device='cuda')
            policy_y, value_y = model(X, requires_grad=False, tensor=False)
            policy_loss, _ = rmse(policy_y, y[:, :-1])
            value_loss, _ = rmse(value_y, y[:, -1])
            policy_value = 2.0
            value_value = 30
            self.assertTrue(policy_loss < policy_value,
                            f'Expected loss = {policy_loss} < {policy_value}')
            self.assertTrue(value_loss < value_value,
                            f'Expected loss = {value_loss} < {value_value}')

        model = GBTModel.load_learner(
            os.path.join(self.test_dir, 'test_cosine_adam_cpu'),
            device='cpu')
        y_pred = model(X, requires_grad=False, tensor=False)
        loss = np.sqrt(np.mean((y_pred.squeeze() - y.squeeze())**2))
        value = 50.0
        self.assertTrue(loss < value, f'Expected loss = {loss} < {value}')


if __name__ == '__main__':
    unittest.main()



##############################################################################
# Copyright (c) 2024, NVIDIA Corporation. All rights reserved.
#
# This work is made available under the Nvidia Source Code License-NC.
# To view a copy of this license, visit
# https://nvlabs.github.io/gbrl/license.html
#
##############################################################################
from typing import Dict, Optional, Tuple

import numpy as np
import torch as th
from gbrl.common.utils import NumericalData

from gbrl.learners.gbt_learner import GBTLearner
from gbrl.models.base import BaseGBT
from gbrl.common.utils import (clip_grad_norm, concatenate_arrays, ensure_leaf_tensor_or_array,
                               numerical_dtype,
                               setup_optimizer, validate_array)


class ContinuousCritic(BaseGBT):
    """
    GBRL model for a Continuous Critic ensemble.
    Designed for Q-function approximation in continuous action spaces,
    such as SAC.
    Model is designed to output parameters of 3 types of Q-functions:
    - linear Q(theta(s), a) = <w_theta, a> + b_theta, (<> denotes a dot product).
    - quadratic Q(theta(s), a) = -(<w_theta, a> - b_theta)**2 + c_theta.
    - tanh Q(theta(s), a) = b_theta*tanh(<w_theta, a>)

    This allows to pass derivatives w.r.t to action a while the Q
    parameters are a function of a GBT model theta.
    The target model is approximated as the ensemble without the last <target_update_interval> trees.
    """
    def __init__(self,
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int,
                 weights_optimizer: Dict,
                 bias_optimizer: Dict,
                 params: Dict = dict(),
                 target_update_interval: int = 100,
                 bias: Optional[np.ndarray] = None,
                 verbose: int = 0,
                 device: str = 'cpu'):
        """
        Initializes the Continuous Critic model.

        Args:
            tree_struct (Dict): Dictionary containing tree structure information:
                    max_depth (int): maximum tree depth.
                    grow_policy (str): 'greedy' or 'oblivious'.
                    n_bins (int): number of bins per feature for candidate generation.
                    min_data_in_leaf (int): minimum number of samples in a leaf.
                    par_th (int): minimum number of samples for parallelizing on CPU.
            output_dim (int): output dimension.
            weights_optimizer Dict: dictionary containing policy optimizer
            parameters. (see GBRL for optimizer details)
            bias_optimizer Dict: dictionary containing policy optimizer parameters.
            (see GBRL for optimizer details)
            params (Dict, optional): GBRL parameters such as:
                control_variates (bool): use control variates (variance reduction technique CPU only).
                split_score_func (str): "cosine" or "l2"
                generator_type- (str): candidate generation method "Quantile" or "Uniform".
                feature_weights - (list[float]): Per-feature multiplication
                weights used when choosing the best split. Weights should be >= 0
            target_update_interval (int): target update interval
            bias (np.ndarray, optional): manually set a bias. Defaults to None = np.zeros.
            verbose (int, optional): verbosity level. Defaults to 0.
            device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """

        self.weights_optimizer = setup_optimizer(weights_optimizer,
                                                 prefix='weights_')
        self.bias_optimizer = setup_optimizer(bias_optimizer, prefix='bias_')

        super().__init__()
        self.target_learner = None
        bias = bias if bias is not None else np.zeros(output_dim, dtype=numerical_dtype)
        self.target_update_interval = target_update_interval
        # init model
        self.learner = GBTLearner(input_dim, output_dim, tree_struct,
                                  [self.weights_optimizer, self.bias_optimizer],
                                  params, verbose, device)
        self.learner.reset()
        self.learner.set_bias(bias)

    def step(self, observations: Optional[NumericalData] = None,
             weight_grad: Optional[NumericalData] = None,
             bias_grad: Optional[NumericalData] = None,
             q_grad_clip: Optional[float] = None) -> None:
        """
        Performs a single boosting step

        Args:
            observations (NumericalData):
            q_grad_clip (float, optional):. Defaults to None.
            weight_grad (Optional[NumericalData], optional): manually calculated gradients. Defaults to None.
            bias_grad (Optional[NumericalData], optional): manually calculated gradients. Defaults to None.
        """
        if observations is None:
            assert self.input is not None, ("Cannot update trees without input."
                                            "Make sure model is called with requires_grad=True")
            observations = self.input
        n_samples = len(observations)
        weight_grad = weight_grad if weight_grad is not None else self.params[0].grad.detach() * n_samples  # type: ignore
        bias_grad = bias_grad if bias_grad is not None else self.params[1].grad.detach() * n_samples  # type: ignore

        weight_grad = clip_grad_norm(weight_grad, q_grad_clip)  # type: ignore
        bias_grad = clip_grad_norm(bias_grad, q_grad_clip)  # type: ignore

        validate_array(weight_grad)
        validate_array(bias_grad)
        theta_grad = concatenate_arrays(weight_grad, bias_grad)

        self.learner.step(observations, theta_grad)  # type: ignore
        self.grad = (weight_grad, bias_grad)  # type: ignore
        self.input = None

    def predict_target(self, observations: NumericalData,
                       tensor: bool = True) -> Tuple[NumericalData,
                                                     NumericalData]:
        """
        Predict the parameters of a Target Continuous Critic as Tensors.
        Prediction is made by summing the outputs the trees from Continuous
        Critic model up to `n_trees - target_update_interval`.

        Args:
            observations (NumericalData):
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Tuple[th.Tensor, th.Tensor]: weights and bias parameters to thetype of Q-functions

        """
        assert self.learner is not None, "Target learner is not set"

        n_trees = self.learner.get_num_trees()
        theta = self.learner.predict(observations, requires_grad=False,
                                     stop_idx=max(n_trees - self.target_update_interval, 1), tensor=tensor)
        weights = theta[:, self.weights_optimizer['start_idx']:self.weights_optimizer['stop_idx']]
        bias = theta[:, self.bias_optimizer['start_idx']:self.bias_optimizer['stop_idx']]

        return weights, bias

    def __call__(self, observations: NumericalData,
                 requires_grad: bool = True, target: bool = False,
                 start_idx: int = 0, stop_idx: Optional[int] = None,
                 tensor: bool = True) -> Tuple[NumericalData,
                                               NumericalData]:
        """
        Predict the parameters of a Continuous Critic as Tensors.
        if `requires_grad=True` then stores ifferentiable parameters in self.params.
           Return type/device is identical to the input type/device.

        Args:
            observations (NumericalData)
            requires_grad (bool, optional). Defaults to True.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction (uses
            all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a numpy array. Defaults to True.

        Returns:
            Tuple[NumericalData, NumericalData]: weights and bias parameters to the type of Q-functions
        """
        if target:
            return self.predict_target(observations, tensor)
        assert self.learner is not None, "Learner model not yet initialized"

        theta = self.learner.predict(observations, requires_grad, start_idx,
                                     stop_idx, tensor)
        weights = theta[:, self.weights_optimizer['start_idx']: self.weights_optimizer['stop_idx']].squeeze()
        bias = theta[:, self.bias_optimizer['start_idx']: self.bias_optimizer['stop_idx']].squeeze()
        weights = ensure_leaf_tensor_or_array(weights, tensor=True,
                                              requires_grad=requires_grad, device=self.learner.device)
        bias = ensure_leaf_tensor_or_array(bias, tensor=True, requires_grad=requires_grad, device=self.learner.device)
        if requires_grad:
            self.grad = None
            self.params = weights, bias  # type: ignore
            self.input = observations
        return weights, bias

    def __copy__(self) -> "ContinuousCritic":
        """
        Creates a copy of the ContinuousCritic model.
        """
        assert self.learner is not None, "Learner model not yet initialized"
        learner = self.learner.copy()
        copy_ = ContinuousCritic(learner.tree_struct, learner.input_dim,
                                 learner.output_dim, learner.optimizers[0],  # type: ignore
                                 learner.optimizers[1], learner.params,  # type: ignore
                                 learner.target_update_interval,  # type: ignore
                                 learner.get.bias(),  # type: ignore
                                 learner.verbose,
                                 learner.device)
        copy_.learner = learner  # type: ignore
        return copy_


class DiscreteCritic(BaseGBT):
    """
    GBRL model for a Discrete Critic ensemble.
    Used for Q-function approximation in discrete action spaces.
    The target model is approximated as the ensemble without the last <target_update_interval> trees.
    """
    def __init__(self,
                 tree_struct: Dict,
                 input_dim: int,
                 output_dim: int,
                 critic_optimizer: Dict,
                 params: Dict = dict(),
                 target_update_interval: int = 100,
                 bias: Optional[np.ndarray] = None,
                 verbose: int = 0,
                 device: str = 'cpu'):
        """
            Initializes the Discrete Critic model.

        Args:
         tree_struct (Dict): Dictionary containing tree structure information:
                max_depth (int): maximum tree depth.
                grow_policy (str): 'greedy' or 'oblivious'.
                n_bins (int): number of bins per feature for candidate generation.
                min_data_in_leaf (int): minimum number of samples in a leaf.
                par_th (int): minimum number of samples for parallelizing on CPU.
        output_dim (int): output dimension.
        critic_optimizer Dict: dictionary containing policy optimizer
        parameters. (see GradientBoostingTrees for optimizer details).
        params (Dict, optional): GBRL parameters such as:
            control_variates (bool): use control variates (variance reduction technique CPU only).
            split_score_func (str): "cosine" or "l2"
            generator_type- (str): candidate generation method "Quantile" or "Uniform".
            feature_weights - (list[float]): Per-feature multiplication
            weights used when choosing the best split. Weights should be >= 0
        verbose (int, optional): verbosity level. Defaults to 0.
        device (str, optional): GBRL device 'cpu' or 'cuda/gpu'. Defaults to 'cpu'.
        """
        critic_optimizer = setup_optimizer(critic_optimizer, prefix='critic_')
        super().__init__()

        self.critic_optimizer = critic_optimizer
        self.target_update_interval = target_update_interval
        bias = bias if bias is not None else np.zeros(output_dim, dtype=numerical_dtype)
        # init model
        self.learner = GBTLearner(input_dim, output_dim, tree_struct,
                                  self.critic_optimizer, params, verbose, device)
        self.learner.reset()
        self.learner.set_bias(bias)

    def step(self, observations: Optional[NumericalData] = None,
             q_grad: Optional[NumericalData] = None,
             max_q_grad_norm: Optional[np.ndarray] = None) -> None:
        """
        Performs a single boosting iterations.

        Args:
            observations (NumericalData):
            max_q_grad_norm (np.ndarray, optional). Defaults to None.
            q_grad (Optional[NumericalData], optional): manually calculated
            gradients. Defaults to None.
        """
        if observations is None:
            assert self.input is not None, ("Cannot update trees without input."
                                            "Make sure model is called with requires_grad=True")
            observations = self.input
        if q_grad is None:
            n_samples = len(observations)
            q_grad = self.params.grad.detach().cpu().numpy() * n_samples  # type: ignore
        q_grad = clip_grad_norm(q_grad, max_q_grad_norm)  # type: ignore

        self.learner.step(observations, q_grad)  # type: ignore
        self.grad = q_grad
        self.input = None

    def __call__(self, observations: NumericalData, requires_grad: bool = True,
                 start_idx: int = 0, stop_idx: Optional[int] = None,
                 tensor: bool = True) -> NumericalData:
        """
        Predict and return Critic's outputs as Tensors. if
        `requires_grad=True` then stores differentiable parameters in self.params.
           Return type/device is identical to the input type/device.

        Args:
            observations (NumericalData)
            requires_grad (bool, optional). Defaults to True.
            start_idx (int, optional): start tree index for prediction. Defaults to 0.
            stop_idx (_type_, optional): stop tree index for prediction
            (uses all trees in the ensemble if set to 0). Defaults to None.
            tensor (bool, optional): Return PyTorch Tensor, False returns a
            numpy array. Defaults to True.

        Returns:
            NumericalData: Critic's outputs.
        """
        assert self.learner is not None, "Learner model not yet initialized"

        q_values = self.learner.predict(observations, requires_grad, start_idx, stop_idx, tensor)
        if requires_grad:
            self.grad = None
            self.params = q_values
            self.input = observations
        return q_values

    def predict_target(self, observations: NumericalData,
                       tensor: bool = True) -> th.Tensor:
        """
        Predict and return Target Critic's outputs as Tensors.
           Prediction is made by summing the outputs the trees from Continuous
           Critic model up to `n_trees - target_update_interval`.

        Args:
            observations (NumericalData)

        Returns:
            th.Tensor: Target Critic's outputs.
        """
        assert self.learner is not None, "Learner model not yet initialized"

        n_trees = self.learner.get_num_trees()
        return self.learner.predict(observations, requires_grad=False,  # type: ignore
                                    stop_idx=max(n_trees - self.target_update_interval, 1),
                                    tensor=tensor)

    def __copy__(self) -> "DiscreteCritic":
        """
        Creates a copy of the DiscreteCritic model.
        """
        assert self.learner is not None, "Learner model not yet initialized"

        learner = self.learner.copy()
        copy_ = DiscreteCritic(learner.tree_struct, learner.input_dim,
                               learner.output_dim, learner.optimizers[0],  # type: ignore
                               learner.params, self.target_update_interval,
                               learner.get_bias(),  # type: ignore
                               learner.verbose,
                               learner.device)
        copy_.learner = learner  # type: ignore
        return copy_
